"""
monobit.encoding.indexers - codepoint indexers

(c) 2022--2024 Rob Hagemans
licence: https://opensource.org/licenses/MIT
"""

from .base import Encoder
from ..core.labels import to_labels


class Indexer(Encoder):
    """Convert from index to ordinals."""

    def __init__(self, code_range='0-'):
        """Index converter."""
        super().__init__('index')
        # generator
        self._code_range = to_labels(code_range)

    @staticmethod
    def char(*labels):
        """Convert codepoint to character, return empty string if missing."""
        raise TypeError('Can only use Indexer to set codepoints, not character labels.')

    def codepoint(self, *labels):
        """Convert character to codepoint."""
        try:
            return next(self._code_range)
        except StopIteration:
            return b''

    def __repr__(self):
        """Representation."""
        return type(self).__name__ + '()'

    @classmethod
    def load(cls, tbl_file):
        """Load indexer from FONTCONV .tbl file"""
        with open(tbl_file) as f:
            tbl = f.read()
        rangestr = '0x' + ',0x'.join(('-0x'.join(tbl.split('-'))).split())
        return cls(code_range=rangestr)



"""
monobit.storage.fontformats.vector.svg - svg writer for vector fonts

(c) 2023--2024 Rob Hagemans
licence: https://opensource.org/licenses/MIT
"""

import re
import logging
from math import ceil
import xml.etree.ElementTree as etree

from monobit.storage import loaders, savers
from monobit.storage.utils.limitations import ensure_single
from monobit.base import Props, reverse_dict, FileFormatError
from monobit.core import Font, Glyph, StrokePath

from ..common import WEIGHT_MAP, WEIGHT_REVERSE_MAP


_STYLE_MAP = {
    'normal': 'roman',
    'italic': 'italic',
    'oblique': 'oblique'
}
_STYLE_REVERSE_MAP = reverse_dict(_STYLE_MAP)

# tag used for 'missing' glyph
DEFAULT_NAME = 'missing'


@loaders.register(
    name='svg',
    patterns=('*.svg',),
    magic=(
        b'<svg>',
        b'<?xml version="1.0" standalone="yes"?>\n<svg'
    ),
)
def load_svg(instream):
    """Load vector font from Scalable Vector Graphics font."""
    root = etree.parse(instream).getroot()
    if not root.tag.endswith('svg'):
        raise FileFormatError(f'Not an SVG file: root tag is {root.tag}')
    # the <font> may optionally be enclosed in a <defs> block
    font = root.find('.//{*}font')
    if not font:
        raise FileFormatError('Not an SVG font file')
    props = Props(
        font_id=font.attrib.get('id'),
    )
    font_face = font.find('{*}font-face')
    if font_face is not None:
        weight = max(100, min(900, int(font_face.attrib.get('font-weight', 400))))
        props |= Props(
            ascent=int(font_face.attrib.get('ascent')),
            descent=-int(font_face.attrib.get('descent')),
            family=font_face.attrib.get('font-family'),
            line_height=int(font_face.attrib.get('font-size', font_face.attrib.get('units-per-em'))),
            underline_thickness=int(font_face.attrib.get('underline-thickness')),
            underline_descent=-int(font_face.attrib.get('underline-position')),
            weight=WEIGHT_MAP[round(weight, -2)],
            slant=_STYLE_MAP.get(font_face.attrib.get('font-style')),
        )
    glyph_elems = list(font.iterfind('{*}glyph'))
    missing_glyph = font.find('{*}missing-glyph')
    if missing_glyph is not None:
        glyph_elems.append(missing_glyph)
    # get the first element containing a path definition
    # either the <glyph> element itself or an enclosed <path>
    # or that path enclosed in <g>s etc
    path_elems = (
        _g.find('.//*[@d]')
        for _g in glyph_elems
    )
    orig_paths = tuple(
        _g.attrib.get('d', '') if _g is not None else ''
        for _g in path_elems
    )
    # convert path to monobit notation
    paths = tuple(convert_path(_p) for _p in orig_paths)
    glyph_props = tuple(
        dict(
            char=_g.attrib.get('unicode', ''),
            advance_width=int(_g.attrib.get('horiz-adv-x', 0)),
            tag=_g.attrib.get('glyph-name', '')
        )
        for _g in glyph_elems
    )
    if missing_glyph is not None:
        glyph_props[-1].update(tag=DEFAULT_NAME)
        props |= Props(default_char=DEFAULT_NAME)
    glyphs = tuple(
        Glyph.from_path(
            _path.shift(0, -props.line_height + props.descent).flip(),
            **_gprop
        )
        for _path, _gprop in zip(paths, glyph_props)
    )
    return Font(glyphs, **vars(props))

def convert_path(svgpath):
    """Convert SVG path to monobit path."""
    # split into individual letters and groups of digits (including minus sign)
    splitgroups = re.compile('[-0-9]+|[a-zA-Z]').findall
    pathit = iter(splitgroups(svgpath))
    x, y = 0, 0
    startx, starty = 0, 0
    path = []
    try:
        for item in pathit:
            if not item:
                continue
            # if it's a number group, the last character must be a digit
            # a number group here mean's we're repeating the last svg path command
            if item[-1].isdigit():
                ds = int(item)
            else:
                svgcommand = item
                ds = None
            if svgcommand in ('m', 'l', 'M', 'L'):
                dx = ds if ds is not None else int(next(pathit))
                dy = int(next(pathit))
                if svgcommand in ('M', 'L'):
                    dx -= x
                    dy -= y
                if svgcommand in ('m', 'M'):
                    command = StrokePath.MOVE
                else:
                    command = StrokePath.LINE
            elif svgcommand in ('h', 'v', 'H', 'V'):
                command = StrokePath.LINE
                if ds is None:
                    ds = int(next(pathit))
                if svgcommand == 'H':
                    ds -= x
                elif svgcommand == 'V':
                    ds -= y
                if svgcommand in ('H', 'h'):
                    dx, dy = ds, 0
                elif svgcommand in ('V', 'v'):
                    dx, dy = 0, ds
            elif svgcommand in ('z', 'Z'):
                # close subpath
                # we asssume that's from the start or the latest move
                command = StrokePath.LINE
                dx, dy = startx - x, starty - y
            else:
                raise ValueError('Curves in paths are not supported.')
            path.append((command, dx, dy))
            x += dx
            y += dy
            if command == StrokePath.MOVE:
                startx, starty = x, y
    except StopIteration:
        logging.warning('Truncated SVG path')
    return StrokePath(path)


def attr_str(attr_dict, indent=0, sep='\n'):
    """Convert a dict to svg element attributes."""
    sep += ' ' * indent
    return sep + sep.join(f'{_k}="{_v}"' for _k, _v in attr_dict.items())


@savers.register(linked=load_svg)
def save_svg(fonts, outfile):
    """Export vector font to Scalable Vector Graphics font."""
    font = ensure_single(fonts)
    # matching whitespace doesn't work as label thinks path-only glyphs are empty
    font = font.label(match_whitespace=False)
    if not any('path' in _g.get_properties() for _g in font.glyphs):
        logging.warning(
            "SVG file will have empty glyphs: no stroke path found"
        )
    outfile = outfile.text
    outfile.write('<svg>\n')
    font_attr = {
        'id': font.font_id or font.family or '0',
        # default advance
        'horiz-adv-x': ceil(font.average_width),
    }
    outfile.write(f'<font{attr_str(font_attr, indent=4)}>\n')
    font_face = {
        'font-family': font.family,
        'units-per-em': font.line_height,
        'font-size': font.line_height,
        'ascent': font.ascent,
        'descent': -font.descent,
        'cap-height': font.cap_height,
        'x-height': font.x_height,
        'underline-thickness': font.underline_thickness,
        'underline-position': -font.underline_descent,
        'font-weight': WEIGHT_REVERSE_MAP.get(font.weight, 400),
        'font-style': _STYLE_REVERSE_MAP.get(font.slant, 'normal'),
    }
    outfile.write(f'  <font-face{attr_str(font_face, indent=6)}/>\n')
    if font.default_char:
        _write_glyph(outfile, font, font.get_default_glyph(), tag='missing-glyph')
    for i, glyph in enumerate(font.glyphs):
        if font.default_char in glyph.tags and len(glyph.get_labels()) == 1:
            # this is *only* the default char, we keep it as missing-glyph
            logging.debug('Skipping default-only glyph `%s`', font.default_char)
            continue
        _write_glyph(outfile, font, glyph)
    outfile.write('</font>\n')
    outfile.write('</svg>\n')


def _write_glyph(outfile, font, glyph, tag='glyph'):
    """Write out a glyph to SVG."""
    if glyph.path:
        path = glyph.path.flip().shift(0, font.line_height-font.descent)
        svgpath = path.as_svg()
        d = f'\n      d="{svgpath}"'
    else:
        d = ''
    charstr = ''.join(f'&#{ord(_c)};' for _c in glyph.char)
    glyphprops = {
        'horiz-adv-x': glyph.advance_width,
    }
    if tag != 'missing-glyph':
        if charstr:
            glyphprops.update({'unicode': charstr})
        if glyph.tags:
            glyphprops.update({'glyph-name': glyph.tags[0]})
    outfile.write(f'  <{tag}{attr_str(glyphprops, indent=0, sep=" ")}>\n')
    outfile.write(f'    <path{d}\n      fill="none" stroke="currentColor" stroke-width="1"/>\n')
    outfile.write(f'  </{tag}>\n')
    # this is shorter but not recognised as single-stroke font by FontForge
    #outfile.write(f'  <{tag}{unicode} horiz-adv-x="{glyph.advance_width}"{d}/>\n')



"""
monobit.storage.fontformats.raw.sixtyfourc - .64C files

(c) 2024 Rob Hagemans
licence: https://opensource.org/licenses/MIT
"""

import logging

from monobit.storage import loaders, savers, Magic
from monobit.core import Glyph, Font, Char
from monobit.base.struct import little_endian as le, bitfield

from .raw import load_bitmap, save_bitmap
from monobit.storage.utils.limitations import ensure_single


###############################################################################
# 64C - two unknown bytes plus 8x8 raw in c64 order (upper- or lowercase)
# I haven't found a sepcification
# large collection of sample files at https://home-2002.code-cop.org/c64/index.html#char

@loaders.register(
    name='64c',
    patterns=('*.64c',),
    # maybe-magic b'\0\x38', b'\0\x20' most comon, but many others occur
)
def load_64c(instream, charset:str='upper'):
    """
    Load a 64C font.

    charset: 'upper' for c64 uppercase & graphical, 'lower' for lowercase & uppercase, '' for not specified
    """
    # the second byte is likely a flag, the first is almost always null
    null = instream.read(1)
    unknown_flags = instream.read(1)
    if null != b'\0':
        logging.warning(f'Non-null first byte %s.', null)
    font = load_bitmap(
        instream,
        width=8, height=8,
    )
    if charset == 'upper':
        font = font.label(char_from='c64')
    elif charset == 'lower':
        font = font.label(char_from='c64-alternate')
    font = font.modify(source_format='64c')
    return font


@savers.register(linked=load_64c)
def save_64c(fonts, outstream):
    """Save a 64C font."""
    font = ensure_single(fonts)
    # not an actual magic sequence. we also see \0\x20 \0\x30 \0\x48 \0\xc8
    outstream.write(b'\x00\x38')
    save_bitmap(outstream, font)



"""
monobit.storage.location - archive, directory and wrapper traversal

(c) 2019--2024 Rob Hagemans
licence: https://opensource.org/licenses/MIT
"""

import logging
import itertools
from pathlib import Path
from collections import deque

from ..plumbing import take_arguments, manage_arguments
from .magic import FileFormatError, MagicRegistry
from .streams import StreamBase, Stream, KeepOpen
from .base import encoders, decoders, containers
from .containers import Container
from .containerformats.directory import Directory


def open_location(
        stream_or_location, mode='r', overwrite=False, match_case=False,
        container_format='', argdict=None, make_dir=False,
    ):
    """Point to given location; may include nested containers and wrappers."""
    if mode not in ('r', 'w'):
        raise ValueError(f"Mode must be 'r' or 'w'; not '{mode}'.")
    if not stream_or_location:
        raise ValueError(f'No location provided.')
    if isinstance(stream_or_location, (str, Path)):
        return Location.from_path(
            stream_or_location, mode=mode,
            overwrite=overwrite, match_case=match_case,
            container_format=container_format, argdict=argdict,
            make_dir=make_dir,
        )
    # assume stream_or_location is a file-like object
    return Location.from_stream(
        stream_or_location, mode=mode,
        overwrite=overwrite, match_case=match_case,
        container_format=container_format, argdict=argdict,
    )


class Location:

    def __init__(
            self, *,
            root=None, path='', mode='r', overwrite=False, match_case=False,
            container_format='', argdict=None, make_dir=False,
        ):
        self.path = Path(path)
        self.mode = mode
        self.overwrite = overwrite
        self.match_case = match_case
        self.resolved = False
        # container or stream on which we attch the path
        # this object is NOT owned by us but externaly provided
        # an further objects in the path will be ours to close
        if isinstance(root, Container):
            self._path_objects = [root]
            self._stream_objects = []
        else:
            self._path_objects = []
            self._stream_objects = [KeepOpen(root)]
        # subpath from last container in path_objects
        self._container_subpath = self.path
        # subpath from last object in path_objects or stream_objects
        self._leafpath = self.path
        # format parameters
        self._container_format = container_format.split('.')
        self._make_dir = make_dir
        self.argdict = argdict
        self._outermost_path = None

    def __repr__(self):
        """String representation."""
        return (
            f"<{type(self).__name__} "
            f"root='{self._path_objects[0]}' path='{self.path}' mode='{self.mode}'"
            f"{' [unresolved]' if not self.resolved else ''}>"
        )

    @classmethod
    def from_path(cls, path, **kwargs):
        """Create from path-like or string."""
        path = Path(path).resolve()
        root = path.anchor
        subpath = path.relative_to(root)
        return cls(
            # Directory objects doesn't really need to be closed
            # so it's OK that we won't close this one
            root=Directory(root),
            path=subpath,
            **kwargs
        )

    @classmethod
    def from_stream(cls, stream, *, subpath='', mode='r', **kwargs):
        """Create from file-like object."""
        if not isinstance(stream, StreamBase):
            # not clear why we need KeepOpen
            # streams mysteriously get closed without it
            # but KeepOpen.close() does not actually get called... :/
            stream = Stream(KeepOpen(stream), mode=mode)
        if stream.mode != mode:
            raise ValueError(
                f"Stream mode '{stream.mode}' not equal to location mode '{mode}'"
            )
        return cls(
            root=stream,
            path=subpath,
            mode=mode,
            **kwargs
        )

    def __enter__(self):
        return self.resolve()

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.close()
        if exc_type == BrokenPipeError:
            return True
        elif exc_type is not None:
            if self._path_objects and self.mode == 'w':
                root = self._path_objects[0]
                if isinstance(root, Directory):
                    root.remove(self._outermost_path)

    def resolve(self):
        """Resolve path, opening streams and containers as needed."""
        self.resolved = True
        try:
            self._resolve()
        except Exception:
            self.close()
            raise
        return self

    def close(self):
        """Close objects we opened on path."""
        # leave out the root object as we don't own it
        while self._stream_objects:
            self._outer_stream_object = self._stream_objects.pop()
            try:
                self._outer_stream_object.close()
            except Exception as exc:
                logging.warning(
                    'Exception while closing %s: %s',
                    self._outer_stream_object, exc
                )
        while len(self._path_objects) > 1:
            outer = self._path_objects.pop()
            try:
                outer.close()
            except Exception as exc:
                logging.warning(
                    'Exception while closing %s: %s',
                    self._outer_stream_object, exc
                )
        self.resolved = False

    @property
    def root(self):
        if self._path_objects:
            return self._path_objects[0]
        else:
            return ''

    @property
    def _leaf(self):
        """Object (stream or container) at the end of path."""
        try:
            return self._stream_objects[-1]
        except IndexError:
            return self._path_objects[-1]

    def get_stream(self):
        """Get open stream at location."""
        if self.is_dir():
            raise IsADirectoryError(f'Location {self} is a directory.')
        stream = self._stream_objects[-1]
        stream.where = self
        return stream

    def is_dir(self):
        """Location points to a directory/container."""
        if not self.resolved:
            raise ValueError(f'Location {self} is not open.')
        return not self._stream_objects

    # directory (container) functionality

    def _get_container_and_subpath(self):
        """Get open container and subpath to location."""
        if not self.is_dir():
            return self._path_objects[-1], self._container_subpath.parent
        return self._path_objects[-1], self._container_subpath

    def join(self, subpath):
        """Get a location at the subpath."""
        assert(self.is_dir())
        return Location(
            root=self._leaf,
            path=self._leafpath / subpath,
            mode=self.mode,
            overwrite=self.overwrite,
            match_case=self.match_case,
        )

    def walk(self):
        """Recursively open locations."""
        if not self.is_dir():
            yield self
            return
        container = self._leaf
        for path in container.iter_sub(self._leafpath):
            subpath = Path(path).relative_to(self._leafpath)
            location = self.join(subpath)
            with location:
                yield from location.walk()

    def iter_sub(self, prefix):
        """List contents of a subpath."""
        container, subpath = self._get_container_and_subpath()
        return (
            _path.relative_to(subpath)
            for _path in container.iter_sub(subpath / prefix)
        )

    def contains(self, item):
        """Check if file is in container. Case sensitive if container/fs is."""
        container, subpath = self._get_container_and_subpath()
        return _contains(container, subpath / item, self.match_case)

    def open(self, name, mode):
        """Open a binary stream in the container."""
        container, subpath = self._get_container_and_subpath()
        self._check_overwrite(container, subpath / name, mode=mode)
        if mode == 'r':
            kwargs = take_arguments(container.decode, self.argdict)
            stream = container.decode(subpath / name, **kwargs)
        else:
            kwargs = take_arguments(container.encode, self.argdict)
            stream = container.encode(subpath / name, **kwargs)
        stream.where = self
        return stream

    def unused_name(self, name):
        """Generate unique name for container file."""
        if not self.contains(name):
            return name
        stem, _, suffix = name.rpartition('.')
        for i in itertools.count():
            filename = '{}.{}'.format(stem, i)
            if suffix:
                filename = '{}.{}'.format(filename, suffix)
            if not self.contains(filename):
                return filename


    ###########################################################################

    def _check_overwrite(self, container, path, mode):
        if (
                mode == 'w' and not self.overwrite
                and _contains(container, path, self.match_case)
            ):
            raise ValueError(
                f"Overwriting existing file '{path}'"
                " requires -overwrite to be set"
            )

    def _resolve(self):
        """
        Convert location to subpath on innermost container and open stream.
        """
        if isinstance(self._leaf, StreamBase):
            self._resolve_wrappers()
        if isinstance(self._leaf, Container):
            self._resolve_subpath()

    def _resolve_wrappers(self):
        """Open one or more wrappers until an unwrapped stream is found."""
        while True:
            stream = self._leaf
            if self._container_format:
                format = self._container_format[-1]
            else:
                format = ''
            try:
                unwrapped_stream = _get_transcoded_stream(
                    stream, mode=self.mode, format=format, argdict=self.argdict,
                )
            except FileFormatError as e:
                # not a wrapper, maybe a container
                logging.debug(e)
                break
            else:
                if self._container_format:
                    self._container_format.pop()
                self._stream_objects.append(unwrapped_stream)
                stream = unwrapped_stream
        # check if innermost stream is a container
        try:
            container_object = _open_container(
                containers, stream, mode=self.mode, format=format,
            )
        except FileFormatError:
            # innermost stream is a non-container stream.
            if self._leafpath == Path('.'):
                return
            raise ValueError(
                f"Cannot open subpath '{self._leafpath}' "
                f"on non-container stream {stream}'"
            )
        else:
            if self._container_format:
                self._container_format.pop()
            self._path_objects.extend(self._stream_objects)
            self._path_objects.append(container_object)
            self._stream_objects = []
            self._container_subpath = self._leafpath

    def _resolve_subpath(self):
        """Resolve subpath on a container object."""
        container = self._leaf
        # stepwise match path elements with existing ones in container
        # head is the innermost existing path element
        head, tail = _match_path(self._leaf, self._leafpath, self.match_case)
        if Path(head) == Path('.') and Path(tail) == Path('.'):
            # path has resolved
            return
        if self.mode == 'r':
            try:
                # see if head points to a file -> open it
                kwargs = take_arguments(container.decode, self.argdict)
                stream = container.decode(head, **kwargs)
            except IsADirectoryError:
                if Path(tail) == Path('.'):
                    # path has resolved; nothing further to open
                    return
                # head (innermost existing) is a subdirectory
                # i.e. tail subpath does not exist
                raise FileNotFoundError(
                    f"Subpath '{tail}' of path '{head}' not found on container {container}."
                )
            else:
                # remove used arguments
                for kwarg in kwargs:
                    del self.argdict[kwarg]
        else:
            # step forward until a suffix is found, or we run out of path
            if not head.suffixes:
                head2, tail = _split_path_suffix(tail)
                head /= head2
            # head is now the innermost path element *to be created*
            # check if we're asked to create an file or a subdirectory
            # it's a subdirectory if (1) explicitly asked or (2) no suffix
            if (
                    self._make_dir
                    or not head.suffixes
                ):
                # head (innermost creatable) should be a subdirectory
                return
            else:
                # head should be a file -> create it
                self._check_overwrite(container, head, mode=self.mode)
                if not self._outermost_path:
                    self._outermost_path = head
                kwargs = take_arguments(container.encode, self.argdict)
                stream = container.encode(head, **kwargs)
                for kwarg in kwargs:
                    del self.argdict[kwarg]
        # recurse on successfully opened file
        self._stream_objects.append(stream)
        self._leafpath = tail
        self._resolve()


def _split_path_suffix(path):
    """Pare forward path until a suffix is found."""
    for head in reversed((path, *path.parents)):
        if head.suffixes:
            tail = path.relative_to(head)
            return head, tail
    # no suffix
    return path, Path('.')


def _step_match(container, matched_path, target, match_case):
    """One-step match for path element."""
    target = str(target)
    for name in container.iter_sub(matched_path):
        found = Path(name).name
        if (found == target) or (
                (not match_case)
                and found.lower() == target.lower()
            ):
            return found
    return ''


def _match_path(container, path, match_case):
    """Stepwise match per path element."""
    segments = Path(path).as_posix().split('/')
    segments = deque(segments)
    matched_path = Path('.')
    while True:
        target = segments.popleft()
        # try case-sensitive match first, then case-insensitive
        match = _step_match(container, matched_path, target, match_case=True)
        if not match and not match_case:
            match = _step_match(container, matched_path, target, match_case=False)
        if match:
            matched_path /= match
            if not segments or not container.is_dir(matched_path):
                # found match this level, can't go deeper
                return matched_path, Path(*segments)
            # found match this level, go to next
            continue
        # no match this level
        return matched_path, Path(target, *segments)


def _contains(container, path, match_case):
    """Container contains file (case insensitive)."""
    _, tail = _match_path(container, path, match_case)
    return tail == Path('.')


def _open_container(
        registry, instream, *,
        format='', mode='r'
    ):
    """Open container on open stream."""
    # identify file type
    fitting_classes = registry.get_for(instream, format=format)
    if not fitting_classes:
        msg = "Container format not recognised"
        if format:
            msg += f': `{format}`'
        raise FileFormatError(msg)
    last_error = None
    for cls in fitting_classes:
        if mode == 'r':
            instream.seek(0)
        logging.info(
            "Opening stream '%s' as container format `%s`",
            instream.name, cls.format
        )
        try:
            # returns container object
            container = cls(instream, mode=mode)
        except FileFormatError as e:
            logging.debug(e)
            last_error = e
            continue
        else:
            return container
    if last_error:
        raise last_error
    message = f"Cannot open container on stream '{instream.name}'"
    if format:
        message += f': format specifier `{format}` not recognised'
    raise FileFormatError(message)


def _get_transcoded_stream(
        instream, *,
        format='', mode='r', argdict=None
    ):
    """Open wrapper on open stream."""
    argdict = argdict or {}
    if mode == 'r':
        registry = decoders
    elif mode == 'w':
        registry = encoders
    else:
        raise ValueError(f"`mode` must be 'r' or 'w', not {mode}.")
    # identify file type
    last_error = None
    for transcoder in iter_funcs_from_registry(registry, instream, format):
        if mode == 'r':
            instream.seek(0)
        logging.info(
            "Transcoding stream '%s' with wrapper format `%s`",
            instream.name, transcoder.format
        )
        try:
            # pick arguments we can use
            kwargs = take_arguments(transcoder, argdict)
            transcoded_stream = transcoder(instream, **kwargs)
        except FileFormatError as e:
            logging.debug(e)
            last_error = e
            continue
        else:
            # remove used arguments
            for kwarg in kwargs:
                del argdict[kwarg]
            return transcoded_stream
    if last_error:
        raise last_error
    message = f"Cannot transcode stream '{instream.name}'"
    if format:
        message += f': format specifier `{format}` not recognised'
    raise FileFormatError(message)


def iter_funcs_from_registry(registry, instream, format):
    """
    Iterate over and wrap functions stored in a MagicRegistry
    that fit a given stream and format.
    """
    # identify file type
    fitting_loaders = registry.get_for(instream, format=format)
    for loader in fitting_loaders:
        yield manage_arguments(loader)
    return



"""
monobit.storage.fontformats.xlfd.pcf - X11 portable compiled format

(c) 2023--2024 Rob Hagemans
licence: https://opensource.org/licenses/MIT
"""

import logging
from itertools import accumulate

from monobit.base.struct import big_endian as be, little_endian as le
from monobit.storage import loaders, savers
from monobit.base import Props
from monobit.core import Font, Glyph, Raster, Tag, Codepoint
from monobit.base.binary import align, ceildiv
from monobit.storage.utils.limitations import ensure_single

from .bdf import swidth_to_pixel, pixel_to_swidth
from .xlfd import (
    parse_xlfd_properties, create_xlfd_properties, create_xlfd_name,
    from_quoted_string
)

MAGIC = b'\1fcp'

@loaders.register(
    name='pcf',
    magic=(MAGIC,),
    patterns=('*.pcf',),
)
def load_pcf(instream):
    """Load font from X11 Portable Compiled Format (PCF)."""
    pcf_data = _read_pcf(instream)
    glyphs = _convert_glyphs(pcf_data)
    props = _convert_props(pcf_data)
    font = Font(glyphs, **props)
    return font.label()

@savers.register(linked=load_pcf)
def save_pcf(
        fonts, outstream, *,
        byte_order:str='big', ink_bounds:bool=True, scan_unit:int=1,
        padding_bytes:int=1, bit_order:str='big',
    ):
    """
    Save font to X11 Portable Compiled Format (PCF).

    ink_bounds: include optional ink-bounds metrics (default: True)
    byte_order: 'big'-endian (default) or 'little'-endian
    bit_order: 'big'-endian (default) if left pixel is most significant bit, or 'little'-endian if left is lsb
    scan_unit: number of bytes per unit in bitmap (1, 2, 4 or 8; default is 1)
    padding_bytes: make raster row a multiple of this number of bytes (1, 2, 4 or 8; default is 1)
    """
    font = ensure_single(fonts)
    _write_pcf(
        outstream, font, endian=byte_order, create_ink_bounds=ink_bounds,
        scan_unit=scan_unit, padding_bytes=padding_bytes, bit_order=bit_order,
    )
    return font


##############################################################################
# https://fontforge.org/docs/techref/pcf-format.html

_HEADER = le.Struct(
    # /* always "\1fcp" */
    header='4s',
    table_count='int32',
)

# fontforge recap has these as apparent signed ints,
# but X sources say CARD32 which is an unsigned int
_TOC_ENTRY = le.Struct(
    # /* See below, indicates which table */
    type='uint32',
    # /* See below, indicates how the data are formatted in the table */
    format='uint32',
    # /* In bytes */
    size='uint32',
    # /* from start of file */
    offset='uint32',
)

# format field
#define PCF_DEFAULT_FORMAT       0x00000000
PCF_DEFAULT_FORMAT = 0x00000000
#define PCF_INKBOUNDS           0x00000200
#define PCF_ACCEL_W_INKBOUNDS   0x00000100
PCF_ACCEL_W_INKBOUNDS = 0x00000100
#define PCF_COMPRESSED_METRICS  0x00000100
PCF_COMPRESSED_METRICS = 0x00000100

# format field modifiers
#define PCF_GLYPH_PAD_MASK       (3<<0)            /* See the bitmap table for explanation */
PCF_GLYPH_PAD_MASK = (3<<0)
#define PCF_BYTE_MASK           (1<<2)            /* If set then Most Sig Byte First */
PCF_BYTE_MASK = (1<<2)
#define PCF_BIT_MASK            (1<<3)            /* If set then Most Sig Bit First */
PCF_BIT_MASK = (1<<3)
#define PCF_SCAN_UNIT_MASK      (3<<4)            /* See the bitmap table for explanation */
PCF_SCAN_UNIT_MASK = (3<<4)

# type field
#define PCF_PROPERTIES               (1<<0)
PCF_PROPERTIES = (1<<0)
#define PCF_ACCELERATORS            (1<<1)
PCF_ACCELERATORS = (1<<1)
#define PCF_METRICS                 (1<<2)
PCF_METRICS = (1<<2)
#define PCF_BITMAPS                 (1<<3)
PCF_BITMAPS = (1<<3)
#define PCF_INK_METRICS             (1<<4)
PCF_INK_METRICS = (1<<4)
#define PCF_BDF_ENCODINGS           (1<<5)
PCF_BDF_ENCODINGS = (1<<5)
#define PCF_SWIDTHS                 (1<<6)
PCF_SWIDTHS = (1<<6)
#define PCF_GLYPH_NAMES             (1<<7)
PCF_GLYPH_NAMES = (1<<7)
#define PCF_BDF_ACCELERATORS        (1<<8)
PCF_BDF_ACCELERATORS = (1<<8)


def _read_pcf(instream):
    """Read font from X11 PCF font file."""
    header = _HEADER.read_from(instream)
    toc = (_TOC_ENTRY * header.table_count).read_from(instream)
    props = Props()
    for entry in toc:
        instream.seek(entry.offset)
        if entry.type == PCF_PROPERTIES:
            props.xlfd_props = _read_properties_table(instream)
        elif entry.type == PCF_ACCELERATORS:
            # mandatory if BDF_ACCELERATORS not defined
            props.acc_props = _read_acc_table(instream)
        elif entry.type == PCF_BDF_ACCELERATORS:
            # optional
            props.bdf_acc_props = _read_acc_table(instream)
        elif entry.type == PCF_METRICS:
            # mandatory
            props.metrics = _read_metrics(instream)
        elif entry.type == PCF_INK_METRICS:
            # optional
            props.ink_metrics = _read_metrics(instream)
        elif entry.type == PCF_BITMAPS:
            # mandatory
            props.bitmap_format, props.bitmaps = _read_bitmaps(instream)
        elif entry.type == PCF_BDF_ENCODINGS:
            # mandatory, but could be empty
            props.encodings, props.default_char = _read_encoding(instream)
        elif entry.type == PCF_SWIDTHS:
            # optional - does not exist in X11 R6.4 sources
            props.swidths = _read_swidths(instream)
        elif entry.type == PCF_GLYPH_NAMES:
            # optional - does not exist in X11 R6.4 sources
            props.glyph_names = _read_glyph_names(instream)
    return props


def _read_format(instream):
    """Read the format record at start of tables."""
    format = int(le.uint32.read_from(instream))
    if format & PCF_BYTE_MASK:
        base = be
    else:
        base = le
    return format, base


# Properties table

# can be be or le
_PROPS = dict(
    name_offset='int32',
    isStringProp='int8',
    value='int32',
)

def _read_properties_table(instream):
    """Read the Properties table."""
    format, base = _read_format(instream)
    nprops = base.uint32.read_from(instream)
    props = (base.Struct(**_PROPS) * nprops).read_from(instream)
    #  pad to next int32 boundary
    padding = instream.read(0 if nprops&3 == 0 else 4-(nprops&3))
    string_size = base.int32.read_from(instream)
    strings = instream.read(string_size)
    xlfd_props = {}
    for prop in props:
        name, _, _ = strings[prop.name_offset:].partition(b'\0')
        name = name.decode('latin-1', 'ignore')
        if prop.isStringProp:
            value, _, _ = strings[int(prop.value):].partition(b'\0')
            value = value.decode('latin-1', 'ignore')
        else:
            value = int(prop.value)
        xlfd_props[name] = value
    return xlfd_props


# Glyph metrics and ink-metrics

# There are two different metrics tables, PCF_METRICS and PCF_INK_METRICS, the
# former contains the size of the stored bitmaps, while the latter contains the
# minimum bounding box. The two may contain the same data, but many CJK fonts
# pad the bitmaps so all bitmaps are the same size.

# from https://tronche.com/gui/x/xlib/graphics/font-metrics/#XCharStruct
# typedef struct {
# 	short lbearing;			/* origin to left edge of raster */
# 	short rbearing;			/* origin to right edge of raster */
# 	short width;			/* advance to next char's origin */
# 	short ascent;			/* baseline to top edge of raster */
# 	short descent;			/* baseline to bottom edge of raster */
# 	unsigned short attributes;	/* per char flags (not predefined) */
# } XCharStruct;

# X docs suggest a right-to-left character would have negative character_width (advance)
# it is unclear about what this means for bearings, as the raster width is given to be rbearing-lbearing
# in any case, in practice RTL chars appear to have positive metrics
# see e.g. cuarabic12.pcf, 6x13-ISO8859-8.pcf
# so we assume character width is always positive, and advance direction is
# decided by the renderer (as we do it)

_UNCOMPRESSED_METRICS = dict(
    left_side_bearing='int16',
    right_side_bearing='int16',
    character_width='int16',
    character_ascent='int16',
    character_descent='int16',
    character_attributes='uint16',
)

# The (compressed) bytes are unsigned bytes which are offset by 0x80
# (so the actual value will be (getc(pcf_file)-0x80). :
_COMPRESSED_METRICS = dict(
    left_side_bearing='uint8',
    right_side_bearing='uint8',
    character_width='uint8',
    character_ascent='uint8',
    character_descent='uint8',
)


# Accelerator table

_ACC_TABLE = dict(
    noOverlap='uint8',
    constantMetrics='uint8',
    terminalFont='uint8',
    constantWidth='uint8',
    inkInside='uint8',
    inkMetrics='uint8',
    drawDirection='uint8',
    padding='uint8',
    fontAscent='int32',
    fontDescent='int32',
    maxOverlap='int32',
    # minimum and maximum value for each metric
    #minbounds=_UNCOMPRESSED_METRICS,
    #maxbounds=_UNCOMPRESSED_METRICS,
    # if format PCF_ACCEL_W_INKBOUNDS:
    # maximum and maximum value for each ink metric
    ##ink_minbounds=_UNCOMPRESSED_METRICS,
    ##ink_maxbounds=_UNCOMPRESSED_METRICS,
)


def _read_acc_table(instream):
    """Read the Accelerator or BDF Accelerator table."""
    format, base = _read_format(instream)
    acc_table = base.Struct(**_ACC_TABLE).read_from(instream)
    acc_table = Props(**vars(acc_table))
    uncompressed_metrics = base.Struct(**_UNCOMPRESSED_METRICS)
    acc_table |= Props(
        minbounds=uncompressed_metrics.read_from(instream),
        maxbounds=uncompressed_metrics.read_from(instream),
    )
    if format & PCF_ACCEL_W_INKBOUNDS:
        acc_table |= Props(
            ink_minbounds=uncompressed_metrics.read_from(instream),
            ink_maxbounds=uncompressed_metrics.read_from(instream),
        )
    return acc_table


def _read_metrics(instream):
    """Read the Metrics or Ink-Metrics table."""
    format, base = _read_format(instream)
    if format & PCF_COMPRESSED_METRICS:
        compressed_metrics = base.Struct(**_COMPRESSED_METRICS)
        # documented as signed int, but unsigned it makes more sense
        # also this is used as uint by bdftopcf for e.g. unifont
        count = base.uint16.read_from(instream)
        metrics = (compressed_metrics * count).read_from(instream)
        # adjust unsigned bytes by 0x80 offset
        metrics = tuple(
            Props(**{_k: _v-0x80 for _k, _v in vars(_m).items()})
            for _m in metrics
        )
    else:
        uncompressed_metrics = base.Struct(**_UNCOMPRESSED_METRICS)
        count = base.uint32.read_from(instream)
        metrics = (uncompressed_metrics * count).read_from(instream)
    return metrics


def _read_bitmaps(instream):
    """Read the Bitmaps table."""
    format, base = _read_format(instream)
    glyph_count = base.int32.read_from(instream)
    offsets = (base.int32 * glyph_count).read_from(instream)
    bitmap_sizes = (base.int32 * 4).read_from(instream)
    bitmap_size = bitmap_sizes[format & 3]
    bitmap_data = instream.read(bitmap_size)
    offsets = tuple(offsets) + (None,)
    return format, tuple(
        bitmap_data[_offs:_next]
        for _offs, _next in zip(offsets, offsets[1:])
    )

# from https://tronche.com/gui/x/xlib/graphics/font-metrics/#XFontStruct
# typedef struct {
# 	XExtData *ext_data;		/* hook for extension to hang data */
# 	Font fid;			/* Font id for this font */
# 	unsigned direction;		/* hint about the direction font is painted */
# 	unsigned min_char_or_byte2;	/* first character */
# 	unsigned max_char_or_byte2;	/* last character */
# 	unsigned min_byte1;		/* first row that exists */
# 	unsigned max_byte1;		/* last row that exists */
# 	Bool all_chars_exist;		/* flag if all characters have nonzero size */
# 	unsigned default_char;		/* char to print for undefined character */
# 	int n_properties;		/* how many properties there are */
# 	XFontProp *properties;		/* pointer to array of additional properties */
# 	XCharStruct min_bounds;		/* minimum bounds over all existing char */
# 	XCharStruct max_bounds;		/* maximum bounds over all existing char */
# 	XCharStruct *per_char;		/* first_char to last_char information */
# 	int ascent;			/* logical extent above baseline for spacing */
# 	int descent;			/* logical decent below baseline for spacing */
# } XFontStruct;

# FontForge docs suggest the encoding table has signed integers
# but the XFontStruct has them unsigned
# they also make more sense unsigned, especially default_char which may be two-byte.
_ENCODING_TABLE = dict(
    min_char_or_byte2='uint16',
    max_char_or_byte2='uint16',
    min_byte1='uint16',
    max_byte1='uint16',
    default_char='uint16',
)

def _generate_codepoints(enc):
    """Generate all code points based on encoding table."""
    if not enc.min_byte1 and not enc.max_byte1:
        codepoints = (
            bytes((_cp,))
            for _cp in range(enc.min_char_or_byte2, enc.max_char_or_byte2+1)
        )
    else:
        codepoints = (
            bytes((_hi, _lo))
            for _hi in range(enc.min_char_or_byte2, enc.max_char_or_byte2+1)
            for _lo in range(enc.min_byte1, enc.max_byte1+1)
        )
    return codepoints

def _read_encoding(instream):
    format, base = _read_format(instream)
    enc = base.Struct(**_ENCODING_TABLE).read_from(instream)
    count = (enc.max_char_or_byte2-enc.min_char_or_byte2+1)*(enc.max_byte1-enc.min_byte1+1)
    # generate code points
    codepoints = _generate_codepoints(enc)
    glyph_indices = (base.int16 * count).read_from(instream)
    encoding_dict = {
        _cp: _idx for _cp, _idx in zip(codepoints, glyph_indices)
        # -1 means 'not used'
        if _idx >= 0
    }
    return encoding_dict, enc.default_char


def _read_swidths(instream):
    """Read the Scalable Widths table."""
    format, base = _read_format(instream)
    glyph_count = base.uint32.read_from(instream)
    swidths = (base.int32 * glyph_count).read_from(instream)
    return swidths


def _read_glyph_names(instream):
    """Read the Glyph Names table."""
    format, base = _read_format(instream)
    glyph_count = base.int32.read_from(instream)
    offsets = (base.int32 * glyph_count).read_from(instream)
    string_size = base.int32.read_from(instream)
    strings = instream.read(string_size)
    glyph_names = []
    for ofs in offsets:
        name, _, _ = strings[ofs:].partition(b'\0')
        name = name.decode('latin-1', 'ignore')
        glyph_names.append(name)
    return glyph_names


###############################################################################
# converter

def _convert_glyphs(pcf_data):
    """Convert glyphs from X11 PCF data to monobit."""
    # label sets
    n_glyphs = len(pcf_data.metrics)
    try:
        # if the table exists, the count should be the same as metrics count
        labelsets = [[Tag(_name)] for _name in pcf_data.glyph_names]
    except AttributeError:
        labelsets = [[] for _ in pcf_data.metrics]
    try:
        for _cp, _idx in pcf_data.encodings.items():
            labelsets[_idx].append(Codepoint(_cp))
    except AttributeError:
        pass
    # scalable width reference values
    if hasattr(pcf_data, 'swidths'):
        dpi_y = pcf_data.xlfd_props.get('RESOLUTION_Y', 72)
        try:
            point_size = pcf_data.xlfd_props['POINT_SIZE'] / 10
        except KeyError:
            try:
                point_size = pcf_data.xlfd_props['PIXEL_SIZE'] * dpi_y / 72 / 10
            except KeyError:
                logging.warning('No point-size information - dropping scalable width table')
                # can't calculate swidths
                pcf_data.swidths = None
    if hasattr(pcf_data, 'swidths') and pcf_data.swidths:
        pcf_data.swidths = tuple(
            swidth_to_pixel(_swidth, point_size, dpi_y)
            for _swidth in pcf_data.swidths
        )
    else:
        pcf_data.swidths = (None,) * n_glyphs
    # /* how each row in each glyph's bitmap is padded (format&3) */
    # /*  0=>bytes, 1=>shorts, 2=>ints */
    glyph_pad_length = pcf_data.bitmap_format & PCF_GLYPH_PAD_MASK
    byte_big = pcf_data.bitmap_format & PCF_BYTE_MASK
    bit_big = pcf_data.bitmap_format & PCF_BIT_MASK
    # /* what the bits are stored in (bytes, shorts, ints) (format>>4)&3 */
    # /*  0=>bytes, 1=>shorts, 2=>ints */
    scan_unit = (pcf_data.bitmap_format & PCF_SCAN_UNIT_MASK) >> 4
    # metrics are as defined in XCharStruct (above)
    # https://tronche.com/gui/x/xlib/graphics/font-metrics/
    # ascent and descent determine the character height, which we instead infer
    # from bitmap and stride.
    glyphs = tuple(
        Glyph.from_bytes(
            _gb,
            width=_met.right_side_bearing-_met.left_side_bearing,
            stride=align(_met.right_side_bearing-_met.left_side_bearing, glyph_pad_length+3),
            height=_met.character_ascent+_met.character_descent,
            bit_order='big' if bit_big else 'little',
            byte_swap=0 if (bool(byte_big) == bool(bit_big)) else 2**scan_unit,
            left_bearing=_met.left_side_bearing,
            right_bearing=_met.character_width-_met.right_side_bearing,
            shift_up=-_met.character_descent,
            scalable_width=_swidth if _swidth != _met.character_width else None,
            labels=_labs,
        )
        for _gb, _met, _labs, _swidth in zip(
            pcf_data.bitmaps, pcf_data.metrics, labelsets, pcf_data.swidths
        )
    )
    return glyphs


def _convert_props(pcf_data):
    """Convert properties for PCF to monobit."""
    xlfd_name = pcf_data.xlfd_props.pop('FONT', '')
    pcf_data.xlfd_props = {_k: str(_v) for _k, _v in pcf_data.xlfd_props.items()}
    props = parse_xlfd_properties(pcf_data.xlfd_props, xlfd_name)
    props.update(dict(
        default_char=Codepoint(pcf_data.default_char),
    ))
    # ascent and descent - these are stored in accelerator table rather than XLFD props
    if hasattr(pcf_data, 'bdf_acc_props'):
        props.update(dict(
            ascent=pcf_data.bdf_acc_props.fontAscent,
            descent=pcf_data.bdf_acc_props.fontDescent,
        ))
    elif hasattr(pcf_data, 'acc_props'):
        props.update(dict(
            ascent=pcf_data.acc_props.fontAscent,
            descent=pcf_data.acc_props.fontDescent,
        ))
    return props


###############################################################################
# pcf writer

def _write_pcf(
        outstream, font,
        endian, create_ink_bounds, scan_unit, padding_bytes, bit_order,
    ):
    """Write font to X11 PCF font file."""
    if endian[:1].lower() == 'b':
        base = be
    else:
        base = le
    # full format needs to be repeated in every table (at least the low byte)
    # or pcf2bdf won't recognise the bitmaps correctly
    format = PCF_DEFAULT_FORMAT
    if bit_order[:1].lower() == 'b':
        format |= PCF_BIT_MASK
    if base == be:
        format |= PCF_BYTE_MASK
    # /* how each row in each glyph's bitmap is padded (format&3) */
    # /*  0=>bytes, 1=>shorts, 2=>ints */
    # do a int(log2())
    format |= (padding_bytes.bit_length()-1) & PCF_GLYPH_PAD_MASK
    # /* what the bits are stored in (bytes, shorts, ints) (format>>4)&3 */
    # /*  0=>bytes, 1=>shorts, 2=>ints */
    format |= ((scan_unit.bit_length()-1) << 4) & PCF_SCAN_UNIT_MASK
    # tables MUST be in this order or pcf2bdf will reject the file
    tables = (
        (PCF_PROPERTIES, *_create_properties_table(font, format, base)),
        (PCF_ACCELERATORS, *_create_acc_table(font, format, base, create_ink_bounds)),
        (PCF_METRICS, *_create_metrics_table(font, format, base, _create_glyph_metrics)),
        # pcf2bdf doesn't read ink metrics, assume they go here?
        (PCF_INK_METRICS, *_create_metrics_table(font, format, base, _create_ink_metrics)),
        (PCF_BITMAPS, *_create_bitmaps(
            font, format, base,
            scan_unit_bytes=scan_unit, padding_bytes=padding_bytes,
            bit_order=bit_order,
        )),
        (PCF_BDF_ENCODINGS, *_create_encoding(font, format, base)),
        (PCF_SWIDTHS, *_create_swidths(font, format, base)),
        (PCF_GLYPH_NAMES, *_create_glyph_names(font, format, base)),
        # we're using the same table for accelerators and BDF accelerators
        # intended difference is unclear
        (PCF_BDF_ACCELERATORS, *_create_acc_table(font, format, base, create_ink_bounds)),
    )
    # calculate offsets, construct ToC
    offset = _HEADER.size + _TOC_ENTRY.size * len(tables)
    toc = []
    for type, table, format in tables:
        toc.append(_TOC_ENTRY(
            type=type,
            format=format,
            size=len(table),
            offset=offset,
        ))
        # align on 2**5==32-bit boundaries
        offset += align(len(table), 5)
    # write out file
    header = _HEADER(header=MAGIC, table_count=len(tables))
    outstream.write(bytes(header))
    for entry in toc:
        outstream.write(bytes(entry))
    for (_, table, _), next in zip(tables, toc[1:] + [None]):
        outstream.write(table)
        # padding
        if next:
            outstream.write(bytes(next.offset - outstream.tell()))



def _create_properties_table(font, format, base):
    """Create the Properties table."""
    propstrings = bytearray()
    xlfd_props = create_xlfd_properties(font)
    xlfd_props['FONT'] = create_xlfd_name(xlfd_props)
    props = []
    props_struct = base.Struct(**_PROPS)
    for key, value in xlfd_props.items():
        prop = props_struct(
            name_offset=len(propstrings),
            isStringProp=isinstance(value, str),
        )
        propstrings += key.encode('ascii', 'replace') + b'\0'
        if prop.isStringProp:
            prop.value = len(propstrings)
            value = from_quoted_string(value)
            propstrings += value.encode('ascii', 'replace') + b'\0'
        else:
            prop.value = int(value)
        props.append(prop)
    table_bytes = (
        bytes(le.uint32(format))
        + bytes(base.uint32(len(props)))
        + bytes((props_struct * len(props))(*props))
        # pad to next int32 boundary
        + bytes(0 if len(props)&3 == 0 else 4-(len(props)&3))
        + bytes(base.uint32(len(propstrings)))
        + bytes(propstrings)
    )
    return table_bytes, format


def _create_glyph_metrics(glyph, base):
    """Convert monobit glyph properties to PCF metrics."""
    metrics = base.Struct(**_UNCOMPRESSED_METRICS)(
        left_side_bearing=glyph.left_bearing,
        right_side_bearing=glyph.left_bearing + glyph.width,
        character_width=glyph.advance_width,
        character_ascent=glyph.height+glyph.shift_up,
        character_descent=-glyph.shift_up,
        character_attributes=0,
    )
    return metrics

def _create_ink_metrics(glyph, base):
    """Create the Ink Metrics table."""
    return _create_glyph_metrics(glyph.reduce(), base)


def _aggregate_metrics(metrics, aggfunc, base):
    """Get minimum or maximum from listo of metrics."""
    return base.Struct(**_UNCOMPRESSED_METRICS)(
        left_side_bearing=aggfunc(_m.left_side_bearing for _m in metrics),
        right_side_bearing=aggfunc(_m.right_side_bearing for _m in metrics),
        character_width=aggfunc(_m.character_width for _m in metrics),
        character_ascent=aggfunc(_m.character_ascent for _m in metrics),
        character_descent=aggfunc(_m.character_descent for _m in metrics),
        character_attributes=0,
    )

# xc/lib/font/util/fontaccel.c
# FontComputeInfoAccelerators(pFontInfo)
#     FontInfoPtr pFontInfo;
# {
#     pFontInfo->noOverlap = FALSE;
#     if (pFontInfo->maxOverlap <= pFontInfo->minbounds.leftSideBearing)
# 	pFontInfo->noOverlap = TRUE;
#
#     if ((pFontInfo->minbounds.ascent == pFontInfo->maxbounds.ascent) &&
# 	    (pFontInfo->minbounds.descent == pFontInfo->maxbounds.descent) &&
# 	    (pFontInfo->minbounds.leftSideBearing ==
# 	     pFontInfo->maxbounds.leftSideBearing) &&
# 	    (pFontInfo->minbounds.rightSideBearing ==
# 	     pFontInfo->maxbounds.rightSideBearing) &&
# 	    (pFontInfo->minbounds.characterWidth ==
# 	     pFontInfo->maxbounds.characterWidth) &&
#       (pFontInfo->minbounds.attributes == pFontInfo->maxbounds.attributes)) {
# 	pFontInfo->constantMetrics = TRUE;
# 	if ((pFontInfo->maxbounds.leftSideBearing == 0) &&
# 		(pFontInfo->maxbounds.rightSideBearing ==
# 		 pFontInfo->maxbounds.characterWidth) &&
# 		(pFontInfo->maxbounds.ascent == pFontInfo->fontAscent) &&
# 		(pFontInfo->maxbounds.descent == pFontInfo->fontDescent))
# 	    pFontInfo->terminalFont = TRUE;
# 	else
# 	    pFontInfo->terminalFont = FALSE;
#     } else {
# 	pFontInfo->constantMetrics = FALSE;
# 	pFontInfo->terminalFont = FALSE;
#     }
#     if (pFontInfo->minbounds.characterWidth == pFontInfo->maxbounds.characterWidth)
# 	pFontInfo->constantWidth = TRUE;
#     else
# 	pFontInfo->constantWidth = FALSE;
#
#     if ((pFontInfo->minbounds.leftSideBearing >= 0) &&
# 	    (pFontInfo->maxOverlap <= 0) &&
# 	    (pFontInfo->minbounds.ascent >= -pFontInfo->fontDescent) &&
# 	    (pFontInfo->maxbounds.ascent <= pFontInfo->fontAscent) &&
# 	    (-pFontInfo->minbounds.descent <= pFontInfo->fontAscent) &&
# 	    (pFontInfo->maxbounds.descent <= pFontInfo->fontDescent))
# 	pFontInfo->inkInside = TRUE;
#     else
# 	pFontInfo->inkInside = FALSE;
# }

def _create_acc_table(font, format, base, create_ink_bounds):
    """Create the Accelerators table."""
    if create_ink_bounds:
        format |= PCF_ACCEL_W_INKBOUNDS
    if not font.glyphs:
        raise ValueError('No glyphs in font.')
    metrics = tuple(_create_glyph_metrics(_g, base) for _g in font.glyphs)
    ink_metrics = tuple(_create_ink_metrics(_g, base) for _g in font.glyphs)
    minbounds = _aggregate_metrics(metrics, min, base)
    maxbounds = _aggregate_metrics(metrics, max, base)
    if create_ink_bounds:
        ink_minbounds = _aggregate_metrics(ink_metrics, min, base)
        ink_maxbounds = _aggregate_metrics(ink_metrics, max, base)
    # based on fontaccel.c code and FontForge's reverse engineered description
    # this is max(left_bearing + width - advance_width) == max(-right_bearing)
    maxOverlap = max(_m.right_side_bearing - _m.character_width for _m in metrics)
    # based on fontaccel.c
    constantMetrics = minbounds == maxbounds
    # checked against bdftopcf and pcf2bdf - these are bdf's FONT_ASCENT and FONT_DESCENT
    # which are kept only here and not in XLFD properties
    fontAscent = font.ascent
    fontDescent = font.descent
    acc_table = base.Struct(**_ACC_TABLE)(
        # /* if for all i, max(metrics[i].rightSideBearing - metrics[i].characterWidth) */
        # /*      <= minbounds.leftSideBearing */
        noOverlap=(maxOverlap <= minbounds.left_side_bearing),
        # /* Means the perchar field of the XFontStruct can be NULL */
        constantMetrics=constantMetrics,
        # /* constantMetrics true and forall characters: */
        # /*      the left side bearing==0 */
        # /*      the right side bearing== the character's width */
        # /*      the character's ascent==the font's ascent */
        # /*      the character's descent==the font's descent */
        terminalFont=(
            constantMetrics
            and (maxbounds.left_side_bearing == 0)
            and (maxbounds.right_side_bearing == maxbounds.character_width)
            and (maxbounds.character_ascent == fontAscent)
            and (maxbounds.character_descent == fontDescent)
        ),
        # /* monospace font like courier */
        constantWidth=(minbounds.character_width == maxbounds.character_width),
        # /* Means that all inked bits are within the rectangle with x between [0,charwidth] */
        # /*  and y between [-descent,ascent]. So no ink overlaps another char when drawing */
        # ===> this seems to imply fontAscent + fontDescent includes leading
        inkInside=(
            (minbounds.left_side_bearing >= 0)
            and (maxOverlap <= 0)
            and (minbounds.character_ascent >= -fontDescent)
            and (maxbounds.character_ascent <= fontAscent)
            and (-minbounds.character_descent <= fontAscent)
            and (maxbounds.character_descent <= fontDescent)
        ),
        # /* true if the ink metrics differ from the metrics somewhere */
        inkMetrics=create_ink_bounds and any(_m != _i for _m, _i in zip(metrics, ink_metrics)),
        # /* 0=>left to right, 1=>right to left */
        # however in practice this is set to 0 even on fonts with RTL glyphs
        # e.g. /usr/share/fonts/arabic24.pcf.gz
        drawDirection=0,
        padding=0,
        fontAscent=fontAscent,
        fontDescent=fontDescent,
        # where set in X11 code?
        maxOverlap=maxOverlap,
        # minbounds=_UNCOMPRESSED_METRICS,
        # maxbounds=_UNCOMPRESSED_METRICS,
    )
    table_bytes = [
        bytes(le.uint32(format)),
        bytes(acc_table),
        bytes(minbounds),
        bytes(maxbounds),
    ]
    if create_ink_bounds:
        table_bytes.extend([
            bytes(ink_minbounds),
            bytes(ink_maxbounds),
    ])
    return b''.join(table_bytes), format


def _create_metrics_table(font, format, base, create_glyph_metrics):
    """Create the Metrics table."""
    # we don't set PCF_COMPRESSED_METRICS
    metrics = tuple(create_glyph_metrics(_g, base) for _g in font.glyphs)
    table_bytes = (
        bytes(le.uint32(format))
        + bytes(base.int32(len(metrics)))
        + b''.join(bytes(_t) for _t in metrics)
    )
    return table_bytes, format


def _create_bitmaps(
        font, format, base,
        scan_unit_bytes=1, padding_bytes=1, bit_order='little',
    ):
    """Create the Bitmaps table."""
    byte_big = base == be
    bit_big = bit_order[:1].lower() == 'b'
    bitmaps = (
        _g.as_bytes(
            # align rows on padding_bytes boundaries
            stride=ceildiv(_g.width, padding_bytes*8) * padding_bytes*8,
            byte_swap=0 if (bool(byte_big) == bool(bit_big)) else scan_unit_bytes,
            bit_order='big' if bit_big else 'little',
        )
        for _g in font.glyphs
    )
    # align full byte sequence on scan_unit boundaries
    bitmaps = tuple(
        _bits.ljust(ceildiv(len(_bits), scan_unit_bytes) * scan_unit_bytes)
        for _bits in bitmaps
    )
    offsets = tuple(accumulate((len(_b) for _b in bitmaps), initial=0))[:-1]
    offsets = (base.int32 * len(bitmaps))(*offsets)
    bitmap_data = b''.join(bitmaps)
    # bytes # shorts # ints #?
    # apparently we do need to calculate all 4
    bitmap_sizes = [
        sum(
            # align full byte sequence on scan_unit boundaries
            ceildiv(
                # align rows on padding_bytes boundaries
                _g.pixels.get_byte_size(stride=ceildiv(_g.width, 8*2**_p) * 8*(2**_p)),
                scan_unit_bytes
            ) * scan_unit_bytes
            for _g in font.glyphs
        )
        for _p in range(4)
    ]
    assert bitmap_sizes[format&3] == len(bitmap_data), f'{bitmap_sizes[format&3]} != {len(bitmap_data)}'
    bitmap_sizes = (base.int32 * 4)(*bitmap_sizes)
    table_bytes = (
        bytes(le.uint32(format))
        + bytes(base.int32(len(offsets)))
        + bytes(offsets)
        + bytes(bitmap_sizes)
        + bitmap_data
    )
    return table_bytes, format


def _create_encoding(font, format, base):
    """Create the Encoding table."""
    font = font.label(codepoint_from=font.encoding)
    enc = base.Struct(**_ENCODING_TABLE)(
        default_char=int(font.get_default_glyph().codepoint or 0)
    )
    codepoints = font.get_codepoints()
    if not codepoints:
        raise ValueError('No storable code points in font.')
    byte_length = len(max(codepoints))
    if byte_length > 2:
        logging.warning(
            'Code points greater than 2 bytes cannot be stored in PCF.'
        )
    elif byte_length == 1:
        enc.min_byte1 = enc.max_byte1 = 0
        enc.min_char_or_byte2 = ord(min(codepoints))
        enc.max_char_or_byte2 = ord(max(codepoints))
    elif byte_length == 2:
        byte2 = [_cp[0] for _cp in codepoints if len(_cp) == 2]
        if any(len(_cp) == 1 for _cp in codepoints):
            byte2.append(0)
        enc.min_char_or_byte2 = min(byte2)
        enc.max_char_or_byte2 = max(byte2)
        byte1 = tuple(_cp[-1] for _cp in codepoints)
        enc.min_byte1 = min(byte1)
        enc.max_byte1 = max(byte1)
    glyph_indices = []
    for cp in _generate_codepoints(enc):
        try:
            index = font.get_index(cp)
        except KeyError:
            # -1 means 'not used'
            index = -1
        glyph_indices.append(index)
    glyph_indices = (base.int16 * len(glyph_indices))(*glyph_indices)
    table_bytes = (
        bytes(le.uint32(format))
        + bytes(enc)
        + bytes(glyph_indices)
    )
    return table_bytes, format


def _create_swidths(font, format, base):
    """Create the Scalable Widths table."""
    swidths = (base.int32 * len(font.glyphs))(*(
        pixel_to_swidth(_g.scalable_width, font.point_size, font.dpi.x)
        for _g in font.glyphs
    ))
    table_bytes = (
        bytes(le.uint32(format))
        + bytes(base.uint32(len(swidths)))
        + bytes(swidths)
    )
    return table_bytes, format


def _create_glyph_names(font, format, base):
    """Create the Glyph Names table."""
    names = tuple(
        _g.tags[0].value.encode('ascii', 'replace') + b'\0'
        if _g.tags else b'glyph%d\0' % (_i,)
        for _i, _g in enumerate(font.glyphs)
    )
    strings = b''.join(_n for _n in names)
    offsets = tuple(accumulate((len(_n) for _n in names), initial=0))[:-1]
    offsets = (base.uint32 * len(names))(*offsets)
    table_bytes = (
        bytes(le.uint32(format))
        + bytes(base.uint32(len(offsets)))
        + bytes(offsets)
        + bytes(base.uint32(len(strings)))
        + strings
    )
    return table_bytes, format



"""
monobit.storage.fontformats.raw.plus3dos - +3DOS header

(c) 2024 Rob Hagemans
licence: https://opensource.org/licenses/MIT
"""

import logging

from monobit.storage import loaders, savers, Magic
from monobit.base import FileFormatError
from monobit.core import Glyph, Font, Char
from monobit.base.struct import little_endian as le, bitfield

from .raw import load_bitmap, save_bitmap
from monobit.storage.utils.limitations import ensure_single, ensure_charcell


###############################################################################
# raw file with +3DOS header
# https://area51.dev/sinclair/spectrum/3dos/fileheader/
# see also https://github.com/oldherl/psftools/blob/master/tools/zxflib.c


_PLUS3DOS_MAGIC = b'PLUS3DOS\x1a'
_PLUS3DOS_HEADER = le.Struct(
    signature='9s',
    # absorb \x1a in signature
    #eof='byte',
    issue='byte',
    version='byte',
    file_size='uint32',
    # +3 BASIC header
    file_type='byte',
    data_length='uint16',
    # > For File Type 3 - CODE, Param 1 is the load address. Param 2 is unused
    param_1='uint16',
    param_2='uint16',
    unused='byte',
    # end of BASIC header
    reserved=le.uint8 * 104,
    checksum='byte',
)


@loaders.register(
    name='plus3dos',
    magic=(_PLUS3DOS_MAGIC,),
)
def load_plus3dos(instream):
    """Load a 768-byte raw font with +3DOS header."""
    header = _PLUS3DOS_HEADER.read_from(instream)
    logging.debug(header)
    if header.signature != _PLUS3DOS_MAGIC:
        raise FileFormatError(
            f'Not a +3DOS file: incorrect signature {header.signature}.'
        )
    if header.file_type != 3 or header.data_length != 768:
        # file type 3 is CODE
        logging.warning(
            '+3DOS file may not be a font file: '
            'file type %d != 3, data length %d != 768',
            header.file_type, header.data_length
    )
    if sum(bytes(header)[:-1]) & 0xff != header.checksum:
        logging.warning('+3DOS checksum failed.')
    font = load_bitmap(instream, width=8, height=8, count=96, first_codepoint=32)
    font = font.modify(source_format='+3DOS')
    font = font.label(char_from='ascii')
    return font


@savers.register(linked=load_plus3dos)
def save_plus3dos(fonts, outstream):
    """Save a 768-byte raw font with +3DOS header."""
    font = ensure_single(fonts)
    header = _PLUS3DOS_HEADER(
        signature=_PLUS3DOS_MAGIC,
        issue=1,
        version=0,
        file_size=768 + _PLUS3DOS_HEADER.size,
        file_type=3,
        data_length=768,
        # following psftools in the choice for these:
        # load address
        param_1=0x3d00,
        unused=ord('f'),
    )
    header.checksum = sum(bytes(header)) & 0xff
    font = font.resample(
        # need explicit Char as to_label does surprising (in this context) things
        # with numeric strings like '1'
        chars=(Char(chr(_c)) for _c in range(32, 128)),
        missing=font.get_glyph(' '),
    )
    outstream.write(bytes(header))
    save_bitmap(outstream, font)



"""
monobit.storage.base - storage

(c) 2024 Rob Hagemans
licence: https://opensource.org/licenses/MIT
"""

from .magic import MagicRegistry

DEFAULT_TEXT_FORMAT = 'yaff'
DEFAULT_BINARY_FORMAT = 'raw'

loaders = MagicRegistry(DEFAULT_TEXT_FORMAT, DEFAULT_BINARY_FORMAT)
savers = MagicRegistry(DEFAULT_TEXT_FORMAT)

container_loaders = MagicRegistry()
container_savers = MagicRegistry()

containers = MagicRegistry()

# wrappers
encoders = MagicRegistry()
decoders = MagicRegistry()



"""
monobit.storage.fontformats.riscos - Acorn RiscOS fonts

(c) 2024 Rob Hagemans
licence: https://opensource.org/licenses/MIT
"""

import logging
from itertools import chain

from monobit.base.binary import ceildiv
from monobit.base.properties import Props
from monobit.base.struct import little_endian as le, bitfield, flag
from monobit.storage import loaders, Magic, Regex
from monobit.base import FileFormatError, UnsupportedError
from monobit.core import Font, Glyph, Raster

from .pkfont import unpack_bits

# https://web.archive.org/web/20210610120924/https://hwiegman.home.xs4all.nl/fileformats/acorn_font/font.htm
# > (The original information can be found on page 1801 of Volume IV of
# > the RISC OS 3.1 PROGRAMMER'S REFERENCE MANUAL)

# extended format versions exist, not yet supported


# x90y45 old-style bitmap file

_INDEX_ENTRY = le.Struct(
    # > 1       point size (not multiplied by 16)
    point_size='uint8',
    # > 1       bits per pixel (4)
    bpp='uint8',
    # > 1       pixels per inch (x-direction)
    dpi_x='uint8',
    # > 1       pixels per inch (y-direction)
    dpi_y='uint8',
    # > 4       reserved - currently 0
    reserved='uint32',
    # > 4       offset of font data in file
    offset='uint32',
    # > 4       size of font data
    size='uint32',
)

_FONT_DATA = le.Struct(
    # > 4       x-size in 1/16ths point * x pixels per inch
    x_size='uint32',
    # > 4       y-size in 1/16ths point * y pixels per inch
    y_size='uint32',
    # > 4       pixels per inch in the x-direction
    dpi_x='uint32',
    # > 4       pixels per inch in the y-direction
    dpi_y='uint32',
    # > 1       x0     |  maximum bounding box for any character
    x0='int8',
    # > 1       y0     |  bottom-left is inclusive
    y0='int8',
    # > 1       x1     |  top-rightis exclusive
    x1='int8',
    # > 1       y1     |  all coordinates are in pixels
    y1='int8',
    # > 512     2 byte offsets from table start of character data. A zero value
    # >         means the character is not defined. These are low/high byte
    # >         pairs (ie little-endian)
    offsets=le.uint16 * 256,
)

_CHAR_HEADER = le.Struct(
    # > 1       x0     |  bounding box
    x0='int8',
    # > 1       y0     |
    y0='int8',
    # > 1       x1 - x0 = X
    width='int8',
    # > 1       y1 - y0 = Y
    height='int8',
    # > X*Y/2   4-bits per pixel (bpp), consectutive rows bottom to top.
    # >         Not aligned until at the end.
    # > 0 - 3.5 Alignment
)

@loaders.register(
    name='riscos-xy',
    # maybe, assumes 4bpp, xdpi 90, ydpi 45 in first entry
    # followed by four nulls in newer but not older files
    magic=(Magic.offset(1) + b'\x04\x5a\x2d',),
    patterns=('x90y45',),
)
def load_x90y45(instream):
    """Load font from acorn RiscOS x90y45 font files."""
    header, metrics = _load_metrics_if_found(instream)
    index = []
    while True:
        entry = _INDEX_ENTRY.read_from(instream)
        # list terminated by a single null
        # so we've read too far, but will seek to the offset anyway
        if not entry.point_size:
            break
        index.append(entry)
        logging.debug('index entry: %s', entry)
    font_data = []
    char_tables = []
    for entry in index:
        instream.seek(entry.offset)
        font_entry = _FONT_DATA.read_from(instream)
        font_data.append(font_entry)
        logging.debug('font data: %s', font_entry)
        # actually offsets are from start of offset table
        anchor = instream.tell() - 512
        char_table = []
        for cp, offset in enumerate(font_entry.offsets):
            if offset == 0:
                continue
            if metrics:
                metrics_index = header.mapping[cp]
                char_metrics = metrics[metrics_index]
            else:
                char_metrics = None
            instream.seek(anchor + offset)
            char_entry = _CHAR_HEADER.read_from(instream)
            char_data = instream.read(ceildiv(char_entry.width * char_entry.height, 2))
            char_table.append((cp, char_entry, char_data, char_metrics))
        char_tables.append(char_table)
    # convert
    fonts = []
    for index_entry, font_entry, char_table in zip(index, font_data, char_tables):
        # entries are in 1/1000 of an em (1em == point-size)
        # point = 1/72 in, 90dpi -> 90 pixels per inch -> 90/72 pixels per point
        scale_x = index_entry.point_size * font_entry.dpi_x / 72 / 1000
        scale_y = index_entry.point_size * font_entry.dpi_y / 72 / 1000
        glyphs = tuple(
            Glyph(
                Raster.from_bytes(
                    char_data,
                    width=char_entry.width,
                    height=char_entry.height,
                    # bpp must be 4
                    bits_per_pixel=index_entry.bpp,
                    align='bit',
                    bit_order='little',
                    bytes=len(char_data),
                ).flip(),
                # x0, y0 in the x90y45 file seem to make more sense than in IntMetrics
                # or I am not understanding the units correctly
                shift_up=char_entry.y0,
                left_bearing=char_entry.x0,
                # y1 seems to be more consistent than y0
                # shift_up=round(char_metrics.y1*scale_y) - char_entry.height + 1,
                # left_bearing=round(char_metrics.x0*scale_x),
                # right bearing such that advance width is consistent
                right_bearing=(
                    round(char_metrics.x_offset*scale_x)
                    - char_entry.x0
                    - char_entry.width
                ) if char_metrics else None,
                scalable_width=(
                    char_metrics.x_offset * scale_x if char_metrics else None
                ),
                codepoint=cp,
            )
            for cp, char_entry, char_data, char_metrics in char_table
        )
        if header:
            name = header.name.rstrip(b'\r').decode('latin-1')
            family, _, subfamily = name.partition('.')
        else:
            family, subfamily = None, None
        fonts.append(Font(
            glyphs,
            family=family,
            subfamily=subfamily or None,
            point_size=index_entry.point_size,
            dpi=(font_entry.dpi_x, font_entry.dpi_y),
            encoding='latin-1',
        ).label())
    return fonts


# IntMetrics file

_INT_METRICS_HEADER = le.Struct(
    # > 40      Name of font, padded with Return (&0D) characters
    name='40s',
    # > 4       16
    # > 4       16
    four_0='uint32',
    four_1='uint32',
    # > 1       n=number of defined characters
    n='uint8',
    # > 3       reserved - currently 0
    reserved=le.uint8 * 3,
    # > 256     character mapping (ie indices into following arrays). For
    # >         example, if the 40th byte in this block is 4, then the fourth
    # >         entry in each of the following arrays refers to that character.
    # >         A zero entry means that character is not defined in this font.
    mapping=le.uint8 * 256,
    # 2n      x0     |
    # 2n      y0     |  bounding cox of character (in 1/1000th em)
    # 2n      x1     |  coordinates are relative to the 'origin point'
    # 2n      y1     |
    # 2n      x-offset after printing this character
    # 2n      y-offset after printing this character
    #
    # The bounding boxes and offsets are given as 16-bit signed numbers, with the low byte first.
)


def _read_int_metrics(instream):
    """Read IntMetrics file."""
    header = _INT_METRICS_HEADER.read_from(instream)
    x0 = (le.int16 * header.n).read_from(instream)
    y0 = (le.int16 * header.n).read_from(instream)
    x1 = (le.int16 * header.n).read_from(instream)
    y1 = (le.int16 * header.n).read_from(instream)
    x_offset = (le.int16 * header.n).read_from(instream)
    y_offset = (le.int16 * header.n).read_from(instream)
    logging.debug('header: %s', header)
    metrics = tuple(
        Props(x0=_x0, y0=_y0, x1=_x1, y1=_y1, x_offset=_xoffs, y_offset=_yoffs)
        for _x0, _y0, _x1, _y1, _xoffs, _yoffs
        in zip(x0, y0, x1, y1, x_offset, y_offset)
    )
    return header, metrics


def _load_metrics_if_found(instream):
    """Load metrics, if available"""
    location = instream.where
    try:
        metrics_file = location.open('IntMetrics', mode='r')
    except FileNotFoundError:
        header = None
        metrics = None
    else:
        with metrics_file:
            header, metrics = _read_int_metrics(metrics_file)
    return header, metrics


###############################################################################
# new font file format

_NEW_HEADER = le.Struct(
    # > 4 "FONT" -identification word
    identification_word='4s',
    # > 1 Bits per pixel:
    # >   0 = outlines
    # >   1 = 1 bpp
    # >   4 = 4 bpp
    bits_per_pixel='uint8',
    # > 1 Version number of file format
    # >   4: no "don't draw skeleton lines unless smaller than this" byte present
    # >   5: byte at [table+512] =max pixel size for skeleton lines
    # >      (0 => always do it)
    # >   6: byte at [chunk+indexsize] = dependency mask (see below)
    version='uint8',
    # > 2 if bpp = 0: design size of font
    # >   if bpp > 0: flags:
    # >     bit 0 set - horizontal subpixel placement
    # >     bit 1 set - vertical subpixel placement
    flags='uint16',
    # > 2 xO - font bounding box (16- bit signed)
    # > 2 yO - units are pixels or design units
    # > 2 x1-xO : xO, yO inclusive, x1, y1 exclusive
    # > 2 y1-y0
    x0='int16',
    y0='int16',
    x1='int16',
    y1='int16',
    # > 4 file offset of 0...31 chunk (word-aligned)
    # > 4 file offset of 32...63 chunk
    # > ...
    # > 4 file offset of 224...255 chunk
    # > 4 file offset of end (ie. size of file)
    # > if offset(n+1) = offset(n), then chunk n is null.
    offsets=le.uint32*9,
)

_TABLE = le.Struct(
    # > 2 n = size of table/scaffold data
    n='uint16',
    # > Bitmaps: (n=10 normally - other values are reserved)
    # > 2 x-size (1/16th point)
    x_size='uint16',
    # 2 x-res (dpi)
    x_res='uint16',
    # 2 y-size (1/16th point)
    y_size='uint16',
    # 2 y-res (dpi)
    y_res='uint16',
)

_CHAR_FLAGS = le.Struct(
    # > 1   flags:
    # >     bit 0 set => coords are 12-bit, else 8-bit
    # >     bit 1 set => data is 1-bpp, else 4-bpp
    # >     bit 2 set => initial pixel is black, else white
    # >     bit 3 set => data is outline, else bitmap
    # >     bits 4-7 = 'f' value for char (0 ==> not encoded)
    coords_12bit=flag,
    data_1bpp=flag,
    initial_black=flag,
    outline=flag,
    f_value=bitfield('uint8', 4),
)
_CHAR_DATA_8BIT = le.Struct(
    # > 2/3 xO, yO sign-extended 8· or 12- bit coordinates
    x0='int8',
    y0='int8',
    # > 2/3 xs, ys width, height (bbox = xO,yO,xO+xs,yO+ys)
    xs='int8',
    ys='int8',
    # > n   data: (depends on type of file)
    #           1-bpp uncrunched: rows from bottom to top
    #           4-bpp uncrunchcd: rows from bottom to top
    #           1-bpp crunched: list of (packed) run-lengths
    #           outlines: list of move/line/curve segments
    # word-aligned at the end of the character data
)
_CHAR_DATA_12BIT = le.Struct(
    # packed pairs of signed 12-bit ints. how does this work if little-endian?
    # it's only specified that "2-byte quantities are little endian"
    x0_y0=le.uint8 * 3,
    xs_ys=le.uint8 * 3,
)

def _from_12bit_pair(x0_y0):
    """Convert 12-bit packed pairs to signed int."""
    # assuming x and y are stored in that order, and 12-bit values are big-endian.
    x0_y0 = int.from_bytes(x0_y0, 'big')
    x0, y0 = divmod(x0_y0, 0x1000)
    # signed values
    if x0 & 0x800:
        x0 = 0x1000 - x0
    if y0 & 0x800:
        y0 = 0x1000 - y0
    return x0, y0


def _convert_12bit_char_data(char_data):
    """Convert 12-bit packed struct to signed int values."""
    x0, y0 = _from_12bit_pair(char_data.x0_y0)
    xs, ys = _from_12bit_pair(char_data.xs_ys)
    return Props(x0=x0, y0=y0, xs=xs, yx=ys)


_RISCOS_MAGIC = b'FONT'

@loaders.register(
    name='riscos',
    # maybe, assumes 4bpp, xdpi 90, ydpi 45 in first entry
    # followed by four nulls in newer but not older files
    magic=(_RISCOS_MAGIC,),
    patterns=(Regex(r'[fb]\d+x\d+'),),
)
def load_riscos(instream):
    """Load font from acorn RiscOS new-style font files."""
    metrics_header, metrics = _load_metrics_if_found(instream)
    header = _NEW_HEADER.read_from(instream)
    logging.debug('header: %s', header)
    if header.identification_word != _RISCOS_MAGIC:
        raise FileFormatError(
            'Not a RiscOS new font format file: magic bytes '
            f'{header.identification_word} != {_RISCOS_MAGIC}'
        )
    if not header.bits_per_pixel:
        raise UnsupportedError('RiscOS Outline fonts not supported')
    table = _TABLE.read_from(instream)
    logging.debug('table: %s', table)
    # NULL-separated description
    desc = instream.read(header.offsets[0] - instream.tell())
    logging.debug('%s', desc)
    glyphs = []
    for chunk_nr, (chunk_start, chunk_end) in enumerate(
            zip(header.offsets[:-1], header.offsets[1:])
        ):
        if chunk_end == chunk_start:
            continue
        instream.seek(chunk_start)
        # > 4 x 32 offset within chunk to character
        # >        0 => character is not defined
        # >        x 4 for vertical placement
        # >        x 4 for horizontal placement
        # Character index is more tightly bound than vertical
        # placement which is more tightly bound than
        # horizontal placement.
        offsets = (le.uint32 * 32).read_from(instream)
        # zero offsets mean character undefined, remove them but keep ordinal
        # we need to know the next offset to be able to determine data length
        index_offsets = tuple(
            (_i, _offs)
            for _i, _offs in enumerate(chain(offsets, (chunk_end-chunk_start,)))
            if _offs != 0
        )
        # ignoring tables for horizontal/vertical subpixel placement
        for (cp, offs), (_, next) in zip(index_offsets[:-1], index_offsets[1:]):
            instream.seek(chunk_start+offs)
            char_flags = _CHAR_FLAGS.read_from(instream)
            if char_flags.coords_12bit:
                char_data = _CHAR_DATA_12BIT.read_from(instream)
                char_data = _convert_12bit_char_data(char_data)
            else:
                char_data = _CHAR_DATA_8BIT.read_from(instream)
            char_bytes = instream.read(chunk_start+next-instream.tell())
            if not char_flags.f_value:
                # f_value == 0 means uncompacted
                # > 1-bpp uncompacted format
                # > 1 bit per pixel, bit set => paint in foreground colour, in
                # > rows from bottom-left to top-right, not aligned until
                # > word-aligned at the end of the character.
                # note the assumptioon is also little-endian in bit order,
                # so the lsb of the first byte maps to the botom left pixel.
                raster = Raster.from_bytes(
                    char_bytes, align='bit', bit_order='little',
                    width=char_data.xs, height=char_data.ys,
                    bits_per_pixel=(1 if char_flags.data_1bpp else 4),
                ).flip()
            else:
                # the packing algorithm is the same as in pkfont
                # except nybbles are taken in little-endian order
                # and pixels are output from bottom left to top right
                bits = unpack_bits(
                    _iter_nybbles_le(char_bytes),
                    char_flags.initial_black, char_flags.f_value, char_data.xs,
                )
                raster = Raster.from_vector(
                    bits, inklevels=(False, True),
                    stride=char_data.xs, width=char_data.xs, height=char_data.ys,
                ).flip()
            codepoint = cp + 32*chunk_nr
            if metrics:
                metrics_index = metrics_header.mapping[codepoint]
                char_metrics = metrics[metrics_index]
            else:
                char_metrics = None
            scale_x = table.x_size * table.x_res / 16 / 72 / 1000
            scale_y = table.y_size * table.y_res / 16 / 72 / 1000
            glyphs.append(Glyph(
                raster, codepoint=codepoint,
                shift_up=char_data.y0,
                left_bearing=char_data.x0,
                right_bearing=(
                    round(char_metrics.x_offset*scale_x)
                    - char_data.x0
                    - char_data.xs
                ) if char_metrics else None,
                scalable_width=(
                    char_metrics.x_offset * scale_x if char_metrics else None
                ),
            ))
    if metrics_header:
        name = metrics_header.name.rstrip(b'\r').decode('latin-1')
        family, _, subfamily = name.partition('.')
    else:
        family, subfamily = None, None
    return Font(
        glyphs,
        family=family,
        subfamily=subfamily or None,
        dpi=(table.x_res, table.y_res),
        point_size=table.y_size/16,
        encoding='latin-1',
        comment='\n'.join(_l for _l in desc.decode('latin-1').split('\0') if _l)
    ).label()


def _iter_nybbles_le(bytestr):
    """Iterate over a bytes string in 4-bit steps (little-endian)."""
    for byte in bytestr:
        hi, lo = divmod(byte, 16)
        yield lo
        yield hi



"""
monobit.storage.fontformats.text.edwin - EDWIN bitmap font format

(c) 2023--2024 Rob Hagemans
licence: https://opensource.org/licenses/MIT
"""

# Edinburgh Drawing-program Which Interacts Nicely
# https://history.dcs.ed.ac.uk/archive/apps/edwin/
# for FNT files see https://gtoal.com/history.dcs.ed.ac.uk/archive/apps/edwin/edwin-apm-August87/
# discussed here https://retrocomputingforum.com/t/hershey-fonts-the-original-vector-fonts/1852/14

import logging
import string

from monobit.storage import loaders, savers
from monobit.base import FileFormatError
from monobit.core import Font, Glyph, Raster
from monobit.base.binary import ceildiv

from monobit.storage.utils.limitations import ensure_single


@loaders.register(
    name='edwin',
    patterns=('*.fnt',),
    text=True,
)
def load_edwin(instream):
    """Load font from EDWIN .FNT file."""
    header = instream.text.readline().strip()
    try:
        first_cp, last_cp = (int(_v) for _v in header.split())
        rasters = []
        for line in instream.text:
            elems = line.strip().split()
            label = elems[0].rstrip(':')
            height, width, *elems = (int(_v) for _v in elems[1:])
            bytestr = b''.join(
                _i.to_bytes(ceildiv(width, 8), 'big')
                for _i in elems
            )
            rasters.append((label, Raster.from_bytes(
                bytestr, height=height, width=width, align='right',
            ).mirror()))
    except ValueError as e:
        raise FileFormatError('Not a well-formed EDWIN file.') from e
    max_height = max(_r.height for _, _r in rasters)
    glyphs = (
        Glyph(
            _raster, shift_up=max_height-_raster.height,
            codepoint=_cp, tag=_label if not _label[:1].isdigit() else None,
        )
        for _cp, (_label, _raster) in enumerate(rasters, first_cp)
    )
    return Font(glyphs)


@savers.register(linked=load_edwin)
def save_edwin(fonts, outstream):
    outstream = outstream.text
    font = ensure_single(fonts)
    # we can only store ascii range
    font = font.subset(chars=tuple(chr(_b) for _b in range(0x80)))
    min_cp, max_cp = int(min(font.get_codepoints())), int(max(font.get_codepoints()))
    # make contiguous
    font = font.resample(codepoints=range(min_cp, max_cp+1))
    font = font.equalise_horizontal()
    outstream.write(f'{min_cp} {max_cp}\n')
    for glyph in font.glyphs:
        rows = (
            str(int.from_bytes(_r, 'big'))
            for _r in glyph.mirror().as_byterows(align='right')
        )
        outstream.write(
            f'{int(glyph.codepoint)}: {glyph.height} {glyph.width} '
        )
        outstream.write(' '.join(rows))
        outstream.write('\n')



"""
monobit.storage.containerformats.sourcecoded - binary files embedded in C/Python/JS source files

(c) 2019--2024 Rob Hagemans
licence: https://opensource.org/licenses/MIT
"""

import string
import logging
from pathlib import Path

from monobit.base import FileFormatError
from ..base import containers
from ..containers import SerialTextContainer
from ..utils.source import (
    CCodeReader, CCodeWriter, PythonCodeReader, PythonCodeWriter,
    PythonTupleCodeReader, PythonTupleCodeWriter,
    JSONCodeReader, JSONCodeWriter,
    PascalCodeReader, PascalCodeWriter,
)

class _CodedBinaryContainer(SerialTextContainer):

    def decode(self, name):
        """
        Decode files from source code file.
        """
        return super().decode(name)

    def encode(self, name, *, bytes_per_line:int=16):
        """
        Encode files to source code file.

        bytes_per_line: number of bytes to write to one line. (default: 16)
        """
        return super().encode(name, bytes_per_line=bytes_per_line)


    ###########################################################################
    # reader

    @classmethod
    def decode_all(cls, instream):
        """Generator to decode all identifiers with payload."""
        instream = instream.text
        start, end = cls.reader.delimiters
        identifier = ''
        data = {}
        for line in instream:
            line = cls.reader.strip_line_comments(line, instream)
            if cls.reader.assign in line:
                identifier, _, _ = line.partition(cls.reader.assign)
                logging.debug('Found assignment to `%s`', identifier)
                identifier = cls.reader.clean_identifier(identifier)
            if identifier and start in line:
                coded_data = cls.reader.read_array(instream, line)
                decoded_data = cls.reader.decode_array(coded_data)
                data[identifier] = decoded_data
                if not decoded_data:
                    logging.warning(
                        "Could not decode data for identifier '%s'",
                        identifier
                    )
                identifier = ''
        return data


    ###############################################################################
    # writer

    @classmethod
    def _head(cls):
        return cls.writer.pre

    @classmethod
    def _tail(cls):
        return cls.writer.post

    @classmethod
    def _separator(cls):
        return cls.writer.separator

    @classmethod
    def _encode(cls, name, rawbytes, outstream, *, bytes_per_line):
        # remove non-ascii
        identifier = name.encode('ascii', 'ignore').decode('ascii')
        identifier = cls.writer.to_identifier(identifier)
        assignment = cls.writer.assign_template.format(
            identifier=identifier, bytesize=len(rawbytes)
        )
        array = cls.writer.encode_array(rawbytes, bytes_per_line)
        outstream.write(assignment)
        outstream.write(array)


@containers.register(
    name='c',
    # don't provide patterns to avoid capturing other source-coded formats
    # patterns=('*.c', '*.cc', '*.cpp', '*.h')
)
class CCodedBinary(_CodedBinaryContainer):
    """C source code wrapper."""
    reader = CCodeReader
    writer = CCodeWriter


@containers.register(
    name='json',
    # patterns=('*.js', '*.json',),
)
class JSONCodedBinary(_CodedBinaryContainer):
    """JSON wrapper."""
    reader = JSONCodeReader
    writer = JSONCodeWriter


@containers.register(
    name='python',
    patterns=('*.py',),
)
class PythonCodedBinary(_CodedBinaryContainer):
    """Python source code wrapper, using lists."""
    reader = PythonCodeReader
    writer = PythonCodeWriter


@containers.register(
    name='python-tuple',
    # patterns=('*.py',),
)
class PythonTupleCodedBinary(_CodedBinaryContainer):
    """Python source code wrapper, using tuples."""
    reader = PythonTupleCodeReader
    writer = PythonTupleCodeWriter


@containers.register(
    name='pascal',
    # patterns=('*.pas',),
)
class PascalCodedBinary(_CodedBinaryContainer):
    """Pascal source code wrapper."""
    reader = PascalCodeReader
    writer = PascalCodeWriter



"""
monobit.render - render to bitmaps

(c) 2019--2024 Rob Hagemans
licence: https://opensource.org/licenses/MIT
"""

from .renderer import render
from .chart import create_chart, grid_map, grid_traverser
from .glyphmap import GlyphMap, glyph_to_image

from . import pdf



"""
monobit.storage.fontformats.common - tables common to various formats

(c) 2019--2024 Rob Hagemans
licence: https://opensource.org/licenses/MIT
"""

from monobit.base import reverse_dict
from monobit.encoding import EncodingName


###############################################################################
# Windows charset values

# official but vague documentation:
# https://docs.microsoft.com/en-us/openspecs/windows_protocols/ms-wmf/0d0b32ac-a836-4bd2-a112-b6000a1b4fc9
#
# The CharacterSet Enumeration defines the possible sets of character glyphs that are defined in fonts for graphics output.
#      typedef  enum
#      {
#        ANSI_CHARSET = 0x00000000,
#        DEFAULT_CHARSET = 0x00000001,
#        SYMBOL_CHARSET = 0x00000002,
#        MAC_CHARSET = 0x0000004D,
#        SHIFTJIS_CHARSET = 0x00000080,
#        HANGUL_CHARSET = 0x00000081,
#        JOHAB_CHARSET = 0x00000082,
#        GB2312_CHARSET = 0x00000086,
#        CHINESEBIG5_CHARSET = 0x00000088,
#        GREEK_CHARSET = 0x000000A1,
#        TURKISH_CHARSET = 0x000000A2,
#        VIETNAMESE_CHARSET = 0x000000A3,
#        HEBREW_CHARSET = 0x000000B1,
#        ARABIC_CHARSET = 0x000000B2,
#        BALTIC_CHARSET = 0x000000BA,
#        RUSSIAN_CHARSET = 0x000000CC,
#        THAI_CHARSET = 0x000000DE,
#        EASTEUROPE_CHARSET = 0x000000EE,
#        OEM_CHARSET = 0x000000FF
#      } CharacterSet;
#
# ANSI_CHARSET: Specifies the English character set.
# DEFAULT_CHARSET: Specifies a character set based on the current system locale; for example, when the system locale is United States English, the default character set is ANSI_CHARSET.
# SYMBOL_CHARSET: Specifies a character set of symbols.
# MAC_CHARSET: Specifies the Apple Macintosh character set.<6>
# SHIFTJIS_CHARSET: Specifies the Japanese character set.
# HANGUL_CHARSET: Also spelled "Hangeul". Specifies the Hangul Korean character set.
# JOHAB_CHARSET: Also spelled "Johap". Specifies the Johab Korean character set.
# GB2312_CHARSET: Specifies the "simplified" Chinese character set for People's Republic of China.
# CHINESEBIG5_CHARSET: Specifies the "traditional" Chinese character set, used mostly in Taiwan and in the Hong Kong and Macao Special Administrative Regions.
# GREEK_CHARSET: Specifies the Greek character set.
# TURKISH_CHARSET: Specifies the Turkish character set.
# VIETNAMESE_CHARSET: Specifies the Vietnamese character set.
# HEBREW_CHARSET: Specifies the Hebrew character set
# ARABIC_CHARSET: Specifies the Arabic character set
# BALTIC_CHARSET: Specifies the Baltic (Northeastern European) character set
# RUSSIAN_CHARSET: Specifies the Russian Cyrillic character set.
# THAI_CHARSET: Specifies the Thai character set.
# EASTEUROPE_CHARSET: Specifies a Eastern European character set.
# OEM_CHARSET: Specifies a mapping to one of the OEM code pages, according to the current system locale setting.

# MS Windows SDK 1.03 Programmer's reference, Appendix C Font Files, p. 427:
#   "One byte specifying the character set defined by this font. The IBM@ PC hardware font has been
#   assigned the designation 377 octal (FF hexadecimal or 255 decimal)."

# below we follow the more useful info at https://www.freetype.org/freetype2/docs/reference/ft2-winfnt_fonts.html
# from freetype freetype/ftwinfnt.h:
#define FT_WinFNT_ID_CP1252    0
#define FT_WinFNT_ID_DEFAULT   1
#define FT_WinFNT_ID_SYMBOL    2
#define FT_WinFNT_ID_MAC      77
#define FT_WinFNT_ID_CP932   128
#define FT_WinFNT_ID_CP949   129
#define FT_WinFNT_ID_CP1361  130
#define FT_WinFNT_ID_CP936   134
#define FT_WinFNT_ID_CP950   136
#define FT_WinFNT_ID_CP1253  161
#define FT_WinFNT_ID_CP1254  162
#define FT_WinFNT_ID_CP1258  163
#define FT_WinFNT_ID_CP1255  177
#define FT_WinFNT_ID_CP1256  178
#define FT_WinFNT_ID_CP1257  186
#define FT_WinFNT_ID_CP1251  204
#define FT_WinFNT_ID_CP874   222
#define FT_WinFNT_ID_CP1250  238
#define FT_WinFNT_ID_OEM     255
# some of their notes:
#   SYMBOL - There is no known mapping table available.
#   OEM - as opposed to ANSI, denotes the second default codepage that most international versions of Windows have.
#         It is one of the OEM codepages from https://docs.microsoft.com/en-us/windows/desktop/intl/code-page-identifiers
#   DEFAULT	- This is used for font enumeration and font creation as a ‘don't care’ value. Valid font files don't contain this value.
#   Exact mapping tables for the various ‘cpXXXX’ encodings (except for ‘cp1361’) can be found at
#   ‘ftp://ftp.unicode.org/Public/’ in the MAPPINGS/VENDORS/MICSFT/WINDOWS subdirectory.
#   ‘cp1361’ is roughly a superset of MAPPINGS/OBSOLETE/EASTASIA/KSC/JOHAB.TXT.
#
CHARSET_MAP = {
    0x00: 'windows-1252',
    # no codepage
    0x01: '',
    0x02: 'windows-symbol',
    0x4d: 'mac-roman',
    0x80: 'windows-932',
    0x81: 'windows-949',
    0x82: 'windows-1361',
    0x86: 'windows-936',
    0x88: 'windows-950',
    0xa1: 'windows-1253',
    0xa2: 'windows-1254',
    0xa3: 'windows-1258',
    0xb1: 'windows-1255',
    0xb2: 'windows-1256',
    0xba: 'windows-1257',
    0xcc: 'windows-1251',
    0xde: 'windows-874',
    0xee: 'windows-1250',
    # could be any OEM codepage
    0xff: '',
}
CHARSET_REVERSE_MAP = reverse_dict(CHARSET_MAP)
CHARSET_REVERSE_MAP.update({
    # different windows versions used different definitions of windows-1252
    # see https://www.aivosto.com/articles/charsets-codepages-windows.html
    'windows-ansi-1.0': 0x00,
    'windows-ansi-2.0': 0x00,
    'windows-ansi-3.1': 0x00,
    'windows-1252': 0x00,
    # windows-1252 agrees with iso-8859-1 (u+0000--u+00ff) except for controls 0x7F-0x9F
    # furthermore, often the control range 0x00-0x19 is set to IBM graphics in windows-1252
    'latin-1': 0x00,
    'unicode': 0x00,
    # use OEM as fallback for undefined as "valid font files don't contain 0x01"
    '': 0xff,
})
CHARSET_REVERSE_MAP = {
    EncodingName(_k): _v
    for _k, _v in CHARSET_REVERSE_MAP.items()
}


###############################################################################
# Windows weight values

# https://web.archive.org/web/20120215123301/http://support.microsoft.com/kb/65123
# dfWeight: 2 bytes specifying the weight of the characters in the character definition data, on a scale of 1 to 1000.
# A dfWeight of 400 specifies a regular weight.
#
# https://docs.microsoft.com/en-gb/windows/desktop/api/wingdi/ns-wingdi-taglogfonta
# The weight of the font in the range 0 through 1000. For example, 400 is normal and 700 is bold.
# If this value is zero, a default weight is used.
# Value	Weight
# FW_DONTCARE	0
# FW_THIN	100
# FW_EXTRALIGHT	200
# FW_ULTRALIGHT	200
# FW_LIGHT	300
# FW_NORMAL	400
# FW_REGULAR	400
# FW_MEDIUM	500
# FW_SEMIBOLD	600
# FW_DEMIBOLD	600
# FW_BOLD	700
# FW_EXTRABOLD	800
# FW_ULTRABOLD	800
# FW_HEAVY	900
# FW_BLACK	900
WEIGHT_MAP = {
    0: '', # undefined/unknown
    100: 'thin', # bdf 'ultra-light' is windows' 'thin'
    200: 'extra-light', # windows 'ultralight' equals 'extralight'
    300: 'light',
    400: 'regular', # 'regular' is 'normal' but less than 'medium' :/ ... bdf has a semi-light here
    500: 'medium',
    600: 'semi-bold',
    700: 'bold',
    800: 'extra-bold', # windows 'ultrabold' equals 'extrabold'
    900: 'heavy', # bdf 'ultra-bold' is 'heavy'
}
WEIGHT_REVERSE_MAP = reverse_dict(WEIGHT_MAP)


###############################################################################
# Mac encoding values

# based on:
# [1] Apple Technotes (As of 2002)/te/te_02.html
# [2] https://developer.apple.com/library/archive/documentation/mac/Text/Text-367.html#HEADING367-0
MAC_ENCODING = {
    0: 'mac-roman',
    1: 'mac-japanese',
    2: 'mac-trad-chinese',
    3: 'mac-korean',
    4: 'mac-arabic',
    5: 'mac-hebrew',
    6: 'mac-greek',
    7: 'mac-cyrillic', # [1] russian
    # 8: [2] right-to-left symbols
    9: 'mac-devanagari',
    10: 'mac-gurmukhi',
    11: 'mac-gujarati',
    21: 'mac-thai',
    23: 'mac-georgian',
    24: 'mac-armenian',
    25: 'mac-simp-chinese', # [1] maldivian
    29: 'mac-centraleurope', # [1] non-cyrillic slavic
    # I don't have a codepage for the following Apple scripts
    # - as Gurmukhi and Gujarati are ISCII-based
    # perhaps we can infer the other ISCII scripts?
    12: 'mac-oriya',
    13: 'mac-bengali',
    14: 'mac-tamil',
    15: 'mac-telugu',
    16: 'mac-kannada',
    17: 'mac-malayalam',
    18: 'mac-sinhalese',
    19: 'mac-burmese',
    20: 'mac-khmer',
    22: 'mac-laotian',
    26: 'mac-tibetan',
    27: 'mac-mongolian',
    28: 'mac-ethiopic', # [2] == geez
    30: 'mac-vietnamese',
    31: 'mac-sindhi', # [2] == ext-arabic
    #32: [1] [2] 'uninterpreted symbols'
}


###############################################################################
# Mac style values

# interpretation of head.macStyle flags
STYLE_MAP = {
    0: 'bold',
    1: 'italic',
    2: 'underline',
    3: 'outline',
    4: 'shadow',
    5: 'condensed',
    6: 'extended',
}

def mac_style_name(font_style):
    """Get human-readable representation of font style."""
    return ' '.join(
        _tag for _bit, _tag in STYLE_MAP.items() if font_style & (1 << _bit)
    )


###############################################################################
# PostScript name

def to_postscript_name(name):
    """Postscript name must be printable ascii, no [](){}<>/%, max 63 chars."""
    ps_name = ''.join(
        _c if _c.isalnum() and _c.isascii() else '-'
        for _c in name
    )
    ps_name = ps_name[:63]
    # expected to be Title-Cased (at least on Mac, see FontForge code comments)
    return ps_name.title()



"""
monobit.unicode - unicode utilities

(c) 2020--2024 Rob Hagemans
licence: https://opensource.org/licenses/MIT
"""

import unicodedata


###################################################################################################
# character properties

def is_fullwidth(char):
    """Check if a character / grapheme sequence is fullwidth."""
    return any(
        unicodedata.east_asian_width(_c) in ('W', 'F')
        for _c in char
    )

def is_graphical(char):
    """Check if a char has a graphical representation."""
    return any(
        # str.isprintable includes everything but Other (C) and Separator (Z), plus SPACE
        # we keep everything but
        # Other/Control (Cc), Other/Surrogate (Cs), Separator/Line (Zl), Separator/Paragraph (Zp)
        # so we keep all spaces (Zs); PUA (Co); Other/Format (Cf) which has things like SOFT HYPHEN
        # also Not Assigned (Cn) - as unicodedata is not up to date
        # anything excluded will be dropped from our charmaps
        unicodedata.category(_c) not in ('Cc', 'Cs', 'Zl', 'Zp')
        for _c in char
    )


def is_showable(char):
    """Check if a char should be shown in yaff files."""
    return (not char) or char == ' ' or is_graphical(char) and all(
        # we keep everything that is_graphical except PUA, Other/Format, Not Assigned, Space Separator
        # anything excluded will be shown as u+XXXX
        unicodedata.category(_c) not in ('Co', 'Cf', 'Cn', 'Zs')
        for _c in char
    )


def is_printable(char):
    """Check if a char should be printed - nothing ambiguous or unrepresentable in there."""
    return (not char) or is_graphical(char) and all(
        # we keep everything that is_graphical except PUA, Other/Format, Not Assigned
        # anything excluded will be shown as REPLACEMENT CHARACTER in codepage charts
        unicodedata.category(_c) not in ('Co', 'Cf', 'Cn')
        for _c in char
    )


def is_blank(char):
    """Check if a sequence is whitespace or non-graphical."""
    if not char:
        return False
    return all(
        unicodedata.category(_c) == 'Zs' or not is_graphical(_c)
        for _c in char
    )

def is_private_use(char):
    """Check if any char is in the private use area."""
    return any(
        unicodedata.category(_c) == 'Co'
        for _c in char
    )

def unicode_name(char, no_name=''):
    """Unicode registered name."""
    names = []
    for c in char:
        try:
            names.append(unicodedata.name(c))
        except ValueError:
            names.append(no_name)
    return ', '.join(names)



"""
monobit.base.imports - supporting functionds for dynamic imports

(c) 2024 Rob Hagemans
licence: https://opensource.org/licenses/MIT
"""

import sys
import logging
from importlib import import_module
from pathlib import Path
from importlib.resources import files


def import_all(module_name):
    """Import all plugins in directory."""
    module = sys.modules[module_name]
    vars(module).update({
        Path(_file).stem: import_module(
            '.' + Path(_file.name).stem,
            module.__package__
        )
        for _file in files(module_name).iterdir()
        if (
            not _file.name.startswith('_')
            and not _file.name.startswith('.')
        )
    })


def safe_import(module_name, name=None):
    """Wrapper for importing external modules and dealing with their errors."""
    item = None
    try:
        module = import_module(module_name)
    except ImportError as e:
        logging.debug('Could not import module `%s`: %s', module_name, e)
    except Exception as e:
        logging.warning('Error while importing module `%s`: %s', module_name, e)
    else:
        if name:
            item = getattr(module, name)
        else:
            item = module
    return item



"""
monobit.storage.fontformats.text.psf2txt - PSF2TXT format

(c) 2019--2024 Rob Hagemans
licence: https://opensource.org/licenses/MIT
"""

import logging
import string

from monobit.storage.base import loaders, savers
from monobit.base import FileFormatError
from monobit.core import Font, Glyph, Char
from monobit.base import Props

from .draw import NonEmptyBlock, Empty, iter_blocks, equal_firsts, Unparsed
from monobit.storage.utils.limitations import ensure_single, ensure_charcell


###############################################################################
# psf2txt

PSFT_KEYS = {
    'Version',
    'Flags',
    'Length',
    'Width',
    'Height',
}

PSFT_CHAR_KEYS = {
    'Unicode',
}

PSFT_SIG = '%PSF2'


@loaders.register(
    name='psf2txt',
    magic=(PSFT_SIG.encode('ascii'),),
    patterns=('*.txt',),
    text=True,
)
def load_psf2txt(instream):
    """Load font from a psf2txt .txt file."""
    properties, glyphs, comments = _read_psf2txt(instream.text)
    return _convert_psf2txt(properties, glyphs, comments)

@savers.register(linked=load_psf2txt)
def save_psf2txt(fonts, outstream):
    """Save font to a psf2txt .txt file."""
    font = ensure_single(fonts)
    _write_psf2txt(font, outstream)


###############################################################################
# writer

def _write_psf2txt(font, outstream):
    font = ensure_charcell(font)
    font = font.label()
    outstream = outstream.text
    outstream.write(PSFT_SIG)
    outstream.write(f'\nVersion: {font.revision or 0}\n')
    # flag 1 means has-unicode-table
    outstream.write('Flags: 1\n')
    outstream.write(f'Length: {len(font.glyphs)}\n')
    outstream.write(f'Width: {font.cell_size.x}\n')
    outstream.write(f'Height: {font.cell_size.y}\n')
    for i, glyph in enumerate(font.glyphs):
        outstream.write(f'%\n// Character {i}\n')
        outstream.write('Bitmap: ')
        outstream.write(
            glyph.as_text(start='', end=' \\\n        ', inklevels='-#')
            .rstrip(' \\\n        ')
        )
        outstream.write('\n')
        if glyph.char:
            outstream.write('Unicode: ')
            for elem in glyph.char:
                outstream.write(f'[{ord(elem):08x}];')
            outstream.write('\n')


###############################################################################
# reader

def _read_psf2txt(text_stream):
    """Read a psf2txt file into a properties object."""
    if text_stream.readline().strip() != '%PSF2':
        raise FileFormatError('Not a PSF2TXT file.')
    properties = {_k: None for _k in PSFT_KEYS}
    comment = []
    glyphs = []
    current_comment = comment
    current_props = properties
    for block in iter_blocks(text_stream, (PTSeparator, PTComment, PTGlyph, PTLabel, PTProperties, Empty)):
        if isinstance(block, PTSeparator):
            if glyphs:
                glyphs[-1] = glyphs[-1].modify(
                    comment='\n\n'.join(current_comment),
                    **current_props
                )
            else:
                properties.update(current_props)
            current_comment = []
            current_props = {}
        elif isinstance(block, PTComment):
            current_comment.append(block.get_value())
        elif isinstance(block, PTProperties):
            current_props.update(block.get_value())
        elif isinstance(block, PTGlyph):
            glyphs.append(Glyph(block.get_value(), inklevels='-#'))
        elif isinstance(block, PTLabel):
            glyphs[-1] = glyphs[-1].modify(labels=block.get_value())
        elif isinstance(block, Unparsed):
            logging.debug('Unparsed lines: %s', block.get_value())
    if current_comment or current_props:
        glyphs[-1] = glyphs[-1].modify(
            comment='\n\n'.join(current_comment),
            **current_props
        )
    return Props(**properties), glyphs, comment


def _convert_psf2txt(props, glyphs, comment):
    mb_props = dict(
        revision=props.Version,
        # ignore Flags, we don't need the others
    )
    return Font(glyphs, **mb_props, comment='\n\n'.join(comment))


# psf2txt block readers

class PTSeparator(NonEmptyBlock):

    def starts(self, line):
        return line[:1] == '%'


class PTComment(NonEmptyBlock):

    def starts(self, line):
        return line.startswith('//')

    def ends(self, line):
        return not self.starts(line)

    def get_value(self):
        lines = tuple(_l.removeprefix('//') for _l in self.lines)
        if equal_firsts(lines) == ' ':
            lines = (_line[1:] for _line in lines)
        return '\n'.join(lines)


class PTGlyph(NonEmptyBlock):

    def starts(self, line):
        return line.startswith('Bitmap:')

    def ends(self, line):
        return line[:1] not in string.whitespace

    def append(self, line):
        line = line.rstrip('\\').strip()
        if line:
            self.lines.append(line)

    def get_value(self):
        _, _, value =  self.lines[0].partition(':')
        value = value.strip()
        lines = self.lines[1:]
        if value:
            lines = [value] + lines
        lines = tuple(_l.strip() for _l in lines)
        return lines


class PTProperties(NonEmptyBlock):

    def starts(self, line):
        return line[:1] in string.ascii_letters and ':' in line

    def ends(self, line):
        return not self.starts(line)

    def get_value(self):
        return dict((_e.strip() for _e in _l.split(':', 1)) for _l in self.lines)


class PTLabel(NonEmptyBlock):

    def starts(self, line):
        return line.startswith('Unicode:')

    def get_value(self):
        _, _, value =  self.lines[0].partition(':')
        value = value.strip()[1:-2].split('];[')
        return tuple(
            Char(''.join(
                chr(int(_cp, 16)) for _cp in _l.split('+'))
            )
            for _l in value
        )



"""
monobit.storage.containerformats.uuencode - UUEncode format

(c) 2024 Rob Hagemans
licence: https://opensource.org/licenses/MIT

modified from the (deprecated) Python standard library uu module source code,
which carries these notices:

> Copyright 1994 by Lance Ellinghouse
> Cathedral City, California Republic, United States of America.
>                        All Rights Reserved
> Permission to use, copy, modify, and distribute this software and its
> documentation for any purpose and without fee is hereby granted,
> provided that the above copyright notice appear in all copies and that
> both that copyright notice and this permission notice appear in
> supporting documentation, and that the name of Lance Ellinghouse
> not be used in advertising or publicity pertaining to distribution
> of the software without specific, written prior permission.
> LANCE ELLINGHOUSE DISCLAIMS ALL WARRANTIES WITH REGARD TO
> THIS SOFTWARE, INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND
> FITNESS, IN NO EVENT SHALL LANCE ELLINGHOUSE CENTRUM BE LIABLE
> FOR ANY SPECIAL, INDIRECT OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
> WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
> ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT
> OF OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
>
> Modified by Jack Jansen, CWI, July 1995:
> - Use binascii module to do the actual line-by-line conversion
>   between ascii and binary. This results in a 1000-fold speedup. The C
>   version is still 5 times faster, though.
"""

import logging
import binascii
from io import BytesIO
from pathlib import Path

from monobit.base import FileFormatError
from ..magic import Sentinel
from ..base import containers
from ..containers import SerialContainer


@containers.register(
    name='uuencode',
    magic=(Sentinel(b'begin '),),
    # ends with 'end\n'
)
class UUEncodeContainer(SerialContainer):
    """UUEncoded format container."""


    def decode(self, name):
        """
        Decode files from source code file.
        """
        return super().decode(name)

    def encode(
            self, name, *,
            permissions:int=0o666,
            backtick:bool=False,
        ):
        """
        Encode files to source code file.

        permissions: Unix permissions for output file(s)
        backtick: subtitute backtick for space
        """
        return super().encode(name, permissions=permissions, backtick=backtick)


    @classmethod
    def decode_all(cls, in_file):
        output = {}
        while True:
            #
            # Read until a begin is encountered or we've exhausted the file
            #
            while True:
                hdr = in_file.readline()
                if not hdr:
                    if not output:
                        raise FileFormatError(
                            'No valid begin line found in input file'
                        )
                    return output
                if not hdr.startswith(b'begin'):
                    continue
                hdrfields = hdr.split(b' ', 2)
                if len(hdrfields) == 3 and hdrfields[0] == b'begin':
                    try:
                        int(hdrfields[1], 8)
                        break
                    except ValueError:
                        pass
            name = (
                hdrfields[2]
                .rstrip(b' \t\r\n\f')
                .decode('ascii', 'replace')
                .replace('\ufffd', '_')
            )
            #
            # Main decoding loop
            #
            s = in_file.readline()
            data = []
            while s and s.strip(b' \t\r\n\f') != b'end':
                try:
                    data.append(binascii.a2b_uu(s))
                except binascii.Error as v:
                    # Workaround for broken uuencoders by /Fredrik Lundh
                    nbytes = (((s[0]-32) & 63) * 4 + 5) // 3
                    data.append(binascii.a2b_uu(s[:nbytes]))
                s = in_file.readline()
            output[name] = b''.join(data)

    @classmethod
    def _encode(cls, name, filedata, out_file, *, permissions, backtick):
        with BytesIO(filedata) as in_file:
            name = name.encode('ascii', 'replace').replace(b'?', b'_')
            out_file.write(
                (b'begin %o %s\n' % ((permissions & 0o777), name))
            )
            data = in_file.read(45)
            while len(data) > 0:
                out_file.write(binascii.b2a_uu(data, backtick=backtick))
                data = in_file.read(45)
            if backtick:
                out_file.write(b'`\nend\n')
            else:
                out_file.write(b' \nend\n')
        out_file.write(b'\n')



"""
monobit.storage.fontformats.raw.writeon - Hercules Write On!

(c) 2024 Rob Hagemans
licence: https://opensource.org/licenses/MIT
"""

import logging

from monobit.storage import loaders, savers, Magic
from monobit.base import FileFormatError
from monobit.core import Glyph, Font, Char
from monobit.base.struct import little_endian as le, bitfield

from .raw import load_bitmap, save_bitmap
from monobit.storage.utils.limitations import ensure_single, ensure_charcell, make_contiguous


###############################################################################
# Hercules Write On! printing utility
# supplied with Hercules InColor utilities disk
# quotes are from the Write On! README section on the CVF.EXE conversion utility
# see also https://www.seasip.info/Unix/PSF/index.html

_WOF_MAGIC = b'\x45\x53\x01\x00'
_WOF_HEADER = le.Struct(
    magic=le.uint8 * 4,
    # header size
    size='uint16',
    # > Char Width is the width of the characters in the font,
    # > expressed in pixels.  This value must be a multiple of 8.
    char_width='uint16',
    # > Char Height is the height of the characters in the font,
    # > expressed in pixels.  This value must be a multiple of 14.
    char_height='uint16',
    # > Tile Height must always be 14 for Write On!  (When Write On!
    # > displays characters that are taller than 14 dots or wider than
    # > 8 dots, it assembles them from two or more 8 x 14 tiles, mosaic
    # > fashion.)
    tile_height='uint16',
    # > Dest Min Char is the ASCII code of the first character in the
    # > destination font (.WOF) file
    min_char='uint16',
    # > Dest Max Char is the ASCII code of the last character in the
    # > destination font (.WOF)
    max_char='uint16',
    unknown_0='uint16',
    maybe_width_too='uint16',
    unknown_1='uint16',
)

@loaders.register(
    name='writeon',
    magic=(_WOF_MAGIC,),
    patterns=('*.wof',),
)
def load_writeon(instream):
    """Load a Write On! font."""
    header = _WOF_HEADER.read_from(instream)
    if bytes(header.magic) != _WOF_MAGIC:
        raise FileFormatError(
            'Not a Hercules Write On! file: '
            f'incorrect signature {bytes(header.magic)}.'
        )
    font = load_bitmap(
        instream,
        width=header.char_width, height=header.char_height,
        first_codepoint=header.min_char,
        count=header.max_char-header.min_char+1,
    )
    font = font.modify(source_format='Hercules Write On!')
    font = font.label(char_from='ascii')
    return font


@savers.register(linked=load_writeon)
def save_writeon(fonts, outstream):
    """Save a Write On! font."""
    font = ensure_single(fonts)
    font = ensure_charcell(font)
    font = make_contiguous(font, supported_range=range(128), missing='space')
    header = _WOF_HEADER(
        magic=le.uint8.array(4)(*_WOF_MAGIC),
        size=_WOF_HEADER.size,
        char_width=font.cell_size.x,
        char_height=font.cell_size.y,
        # > Tile Height must always be 14 for Write On!
        tile_height=14,
        min_char=int(min(font.get_codepoints())),
        max_char=int(max(font.get_codepoints())),
        maybe_width_too=font.cell_size.x,
    )
    outstream.write(bytes(header))
    save_bitmap(outstream, font)



"""
monobit.storage.fontformats.fon.os2 - read OS/2 GPI bitmap fonts and LX containers

(c) 2023--2024 Rob Hagemans
licence: https://opensource.org/licenses/MIT

this implementation leans heavily on Alexander Taylor's os2font
- http://www.altsan.org/programming/os2font_src.zip
- https://github.com/altsan/os2-gpi-font-tools
"""

from monobit.base import import_all
import_all(__name__)



"""
monobit.storage.containerformats.email - files embedded in email messages

(c) 2024 Rob Hagemans
licence: https://opensource.org/licenses/MIT

Based on Python documentation examples at
https://docs.python.org/3/library/email.examples.html

> Thanks to Matthew Dixon Cowles for the original inspiration and examples.

> © Copyright 2001-2024, Python Software Foundation.
> This page is licensed under the Python Software Foundation License Version 2.
> Examples, recipes, and other code in the documentation are additionally
> licensed under the Zero Clause BSD License.
> See https://docs.python.org/license.html for more information.
"""

import email
from email.message import EmailMessage
from email.policy import SMTP, default
import mimetypes
from pathlib import Path

from monobit.base import FileFormatError
from ..base import containers
from ..containers import FlatFilterContainer


@containers.register(
    name='email',
    patterns=('*.eml', '*.msg'),
)
class EmailContainer(FlatFilterContainer):

    def decode(self, name):
        """
        Decode files from email attachments.
        """
        return super().decode(name)

    def encode(
            self, name, *,
            sender:str='me@here',
            recipient:str='you@there',
            subject:str='Files',
        ):
        """
        Encode files to email attachments. Only the email parameters of the last file wrtten have an effect.

        sender: email From field
        recipient: email To field
        subject: email Subject
        """
        return super().encode(
            name,
            sender=sender,
            recipient=recipient,
            subject=subject,
        )

    @classmethod
    def encode_all(cls, data, outstream):
        if not data:
            return
        # Create the message
        msg = EmailMessage()
        msg.preamble = 'This is a multi-part message in MIME format.\n'
        for filename, filedict in data.items():
            msg['Subject'] = filedict['subject']
            msg['To'] = filedict['recipient']
            msg['From'] = filedict['sender']
            filedata = filedict.pop('outstream').getvalue()
            # Guess the content type based on the file's extension.  Encoding
            # will be ignored, although we should check for simple things like
            # gzip'd or compressed files.
            ctype, encoding = mimetypes.guess_type(filename)
            if ctype is None or encoding is not None:
                # No guess could be made, or the file is encoded (compressed), so
                # use a generic bag-of-bits type.
                ctype = 'application/octet-stream'
            maintype, subtype = ctype.split('/', 1)
            msg.add_attachment(
                filedata, maintype=maintype, subtype=subtype, filename=filename
            )
        # Now send or store the message
        outstream.write(msg.as_bytes(policy=SMTP))

    @classmethod
    def decode_all(cls, instream):
        msg = email.message_from_binary_file(instream, policy=default)
        data = {}
        counter = 1
        for part in msg.walk():
            # multipart/* are just containers
            if part.get_content_maintype() == 'multipart':
                continue
            # Applications should really sanitize the given filename so that an
            # email message can't be used to overwrite important files
            filename = part.get_filename()
            if not filename:
                ext = mimetypes.guess_extension(part.get_content_type())
                if not ext:
                    # Use a generic bag-of-bits extension
                    ext = '.bin'
                filename = f'part-{counter:03d}{ext}'
            counter += 1
            data[filename] = part.get_payload(decode=True)
        return data



"""
monobit.render.pdf - pdf chart output

(c) 2019--2024 Rob Hagemans
licence: https://opensource.org/licenses/MIT
"""

import logging

from monobit.base import safe_import
reportlab = safe_import('reportlab')
if reportlab:
    from reportlab.lib.units import mm
    from reportlab.pdfgen.canvas import Canvas

from monobit.storage import savers
from monobit.base import Coord, RGB
from monobit.base.binary import ceildiv
from monobit.core import Font, Codepoint
from monobit.storage.utils.limitations import ensure_single
from .chart import create_chart, aligns_right


if reportlab:

    @savers.register(
        name='pdf',
        patterns=('*.pdf',),
    )
    def save_pdf(
            fonts, outstream, *,
            glyphs_per_line:int=16,
            lines_per_page:int=16,
            padding:Coord=Coord(3, 3),
            direction:str=None,
            pixel_border:RGB=RGB(127, 127, 127),
            ink:RGB=RGB(0, 0, 0),
            paper:RGB=RGB(255, 255, 255),
            codepoint_range:tuple[Codepoint]=None,
            grid_positioning:bool=False,
            max_labels:int=1,
            page_size:Coord=Coord(210, 297),
            margin:Coord=Coord(25, 25),
            title:str='{name}',
            fill_page:bool=False,
        ):
        """
        Export font to chart in Portable Document Format (PDF).

        glyphs_per_line: number of glyphs in primary direction (default: 16)
        lines_per_page: number of lines in secondary direction (default: 16)
        padding: number of pixels in X,Y direction between glyphs (default: 3x3)
        direction: two-part string such as 'left-to-right top-to-bottom'. Default: font direction.
        pixel_border: colour of lines around pixel squares R,G,B, 0--255. use -1,-1,-1 for no border (default: 127,127,127)
        paper: background colour R,G,B 0--255 (default: 255,255,255)
        ink: full-intensity foreground colour R,G,B 0--255 (default: 0,0,0)
        codepoint_range: range of codepoints to include (includes bounds and undefined codepoints; default: all codepoints)
        grid_positioning: place codepoints on corresponding grid positions, leaving gaps if undefined (default: false)
        max_labels: maximum number of labels to show per glyph (default: 1)
        page_size: page size X,Y in millimetres (default 210x297 for A4)
        margin: margin X,Y in millimetres (default 25x25)
        title: title template, using font properties (default: '{name}')
        fill_page: fill out usable space, ignoring pixel aspect ratio (default: False)
        """
        # construct grid pages
        glyph_map = create_chart(
            fonts,
            glyphs_per_line=glyphs_per_line,
            lines_per_page=lines_per_page,
            direction=direction,
            padding=padding,
            codepoint_range=codepoint_range,
            grid_positioning=grid_positioning,
            margin=Coord(0, 0),
            scale=Coord(1, 1),
            max_labels=max_labels,
        )
        max_sheet, min_x, min_y, max_x, max_y = glyph_map.get_bounds()
        font, *_ = fonts

        # assume A4
        # note mm is a )constant defining number of points in a millimetre
        # 1 point = 1/72 in
        page_x, page_y = page_size.x*mm, page_size.y*mm
        # margins and title position
        title_y = (margin.y // 5)*mm
        margin_x, margin_y = margin.x*mm, margin.y*mm + title_y

        # width and height of usable area, in points
        chart_width = page_x - 2*margin_x
        chart_height = page_y - 2*margin_y - title_y
        # width and height of a pixel, in points
        xpix = chart_width / (max_x - min_x)
        # reserve space for labels at the top (min and max exclude padding)
        ypix = chart_height / (max_y - min_y + max_labels + 1)

        if not fill_page:
            # enforce pixel aspect ratio
            xpix_aspect = (ypix * font.pixel_aspect.x) / font.pixel_aspect.y
            ypix_aspect = (xpix * font.pixel_aspect.y) / font.pixel_aspect.x
            if ypix_aspect <= ypix:
                margin_y += chart_height * (1 - ypix_aspect / ypix)
                chart_height *= ypix_aspect / ypix
                ypix = ypix_aspect
            else:
                margin_x += 0.5 * chart_width * (1 - xpix_aspect / xpix)
                chart_width *= xpix_aspect / xpix
                xpix = xpix_aspect

        # horizontal alignment
        # note that create_chart has equalised glyphs horizontally
        direction = direction or font.direction
        right_align = aligns_right(direction)

        canvas = Canvas(outstream)
        # draw title on first page
        canvas.translate(margin_x, margin_y)
        canvas.setFont('Helvetica-Bold', title_y)
        if right_align:
            canvas.drawRightString(
                chart_width, chart_height + title_y,
                font.format_properties(title),
            )
        else:
            canvas.drawString(
                0, chart_height + title_y,
                font.format_properties(title),
            )
        canvas.translate(-margin_x, -margin_y)

        # draw pages
        for sheet in range(max_sheet+1):
            canvas.setPageSize((page_x, page_y))
            canvas.translate(margin_x, margin_y)
            canvas.setLineWidth(xpix / 10)
            # text is the height of one glyph pixel
            canvas.setFont('Helvetica', ypix)

            # output glyph grid
            for record in glyph_map.get_sheet(sheet):
                # draw glyph
                pixels = record.glyph.as_matrix()
                for y in range(len(pixels)):
                    for x in range(len(pixels[y])):
                        fill = pixels[y][x] / (font.levels-1)
                        fill_rgb = tuple(
                            _i * fill/255 + _p * (1-fill)/255
                            for _i, _p in zip(ink, paper)
                        )
                        if all(_c > 0 for _c in pixel_border):
                            stroke_rgb = tuple(_c / 255 for _c in pixel_border)
                        else:
                            stroke_rgb = fill_rgb
                        canvas.setStrokeColorRGB(*stroke_rgb)
                        canvas.setFillColorRGB(*fill_rgb)
                        canvas.rect(
                            (record.x + x) * xpix,
                            (record.y + record.glyph.height - y - 1) * ypix,
                            xpix, ypix,
                            fill=True,
                        )
                        canvas.setFillColorRGB(0, 0, 0)
            canvas.setStrokeColorRGB(0, 0, 0)
            for label in glyph_map.get_sheet_labels(sheet):
                # draw label
                if label.right_align:
                    canvas.drawRightString(label.x*xpix, label.y*ypix, label.text)
                else:
                    canvas.drawString(label.x*xpix, label.y*ypix, label.text)
            canvas.showPage()
        canvas.save()



"""
monobit.plumbing - command-line and scripting support

(c) 2019--2024 Rob Hagemans
licence: https://opensource.org/licenses/MIT
"""

from .scripting import *
from .args import *
# can't import help due to circular dependency



"""
monobit.storage.utils.source - utilities for reading and writing source code

(c) 2019--2024 Rob Hagemans
licence: https://opensource.org/licenses/MIT
"""

import re
import string
from itertools import zip_longest, cycle, chain

from monobit.base import Props
from monobit.core import Raster


class CodeReader:

    @classmethod
    def strip_line_comments(cls, line, instream):
        """Strip comments. Handles inline but not multiline block comments."""
        if cls.comment:
            line, _, _ = line.partition(cls.comment)
        if cls.block_comment:
            start, end = cls.block_comment
            closed = True
            while start in line:
                before, _, after = line.partition(start)
                _, closed, after = after.partition(end)
                line = before + after
            if not closed:
                for new_line in instream:
                    _, closed, after = new_line.partition(end)
                    if closed:
                        line += after
                        break
        line = line.strip(' \r\n\t')
        return line

    @classmethod
    def read_array(cls, instream, line):
        """Retrieve coded array as list of strings. Flattens nested arrays."""
        start, end = cls.delimiters
        depth = 0
        coded_data = []
        for line in chain((line,), instream):
            line = cls.strip_line_comments(line, instream)
            while start in line:
                pre, line = line.split(start, 1)
                if depth:
                    line = pre + line
                depth += 1
            while end in line:
                line, post = line.split(end, 1)
                depth = max(depth - 1, 0)
                if depth:
                    line += post
            coded_data.append(line)
            if not depth:
                break
        return ''.join(coded_data).split(',')

    @classmethod
    def decode_array(cls, coded_data):
        """Decode coded array to bytes."""
        try:
            return bytes(cls.decode_int(_s) for _s in coded_data if _s.strip())
        except ValueError:
            return b''

    @classmethod
    def clean_identifier(cls, identifier):
        """Clean up identifier found in source code."""
        # take last element separated by whitespace e.g. char foo[123] -> foo[123]
        *_, identifier = identifier.strip().split()
        # strip non-alnum at either end (e.g. "abc" -> abc)s
        identifier = re.sub(r"^\W+|\W+$", "", identifier)
        # take first alphanum part (e.g. name[123 -> name)
        identifier, *_ = re.split(r"\W+", identifier)
        return identifier


###############################################################################
# helper functions for writer

class CodeWriter:

    @classmethod
    def to_identifier(cls, identifier):
        """Convert name to C identifier."""
        identifier = ''.join(_c.lower() if _c.isalnum() else '_' for _c in identifier)
        if not identifier[:1].isalpha():
            identifier = 'data_' + identifier
        return identifier

    @classmethod
    def encode_array(
            cls, data, bytes_per_line, lines_per_block=0, add_visuals=False,
            index=None,
        ):
        """Encode bytes to array."""
        start_delimiter, end_delimiter = cls.delimiters
        outstrs = []
        # emit code
        outstrs.append(f'{start_delimiter}\n')
        # grouper
        args = [iter(data)] * bytes_per_line
        groups = list(zip(*args))
        lines = [
            ', '.join(cls.encode_int(_b) for _b in _group)
            for _group in groups
        ]
        if add_visuals and cls.block_comment:
            visuals = [
                Raster
                    .from_bytes(_group, height=1)
                    .as_text(
                        start=cls.block_comment[0] + ' ',
                        end=' ' + cls.block_comment[1],
                    )
                for _group in groups
            ]
        else:
            visuals = []
        rem = len(data) % bytes_per_line
        if rem:
            lines.append(', '.join(
                cls.encode_int(_b) for _b in data[-rem:])
            )
        for i, (line, visual) in enumerate(zip_longest(lines, visuals)):
            if lines_per_block:
                if (i % lines_per_block == 0):
                    # if i:
                    outstrs.append('\n')
                    if index is not None:
                        outstrs.append(
                            cls.indent + cls.block_comment[0] + ' '
                            + cls.encode_int(index+i//lines_per_block)
                            + ' ' + cls.block_comment[1] + '\n'
                        )
            outstrs.append(f'{cls.indent}{line}')
            if i < len(lines) - 1:
                outstrs.append(',')
            if add_visuals:
                outstrs.append(f'{cls.indent}{visual}')
            outstrs.append('\n')
        outstrs.append(end_delimiter)
        return ''.join(outstrs)


###############################################################################
# C

class CCode:
    """C source code."""
    delimiters = '{}'
    comment = '//'
    assign = '='
    block_comment = ('/*', '*/')
    separator = ';\n\n'

    indent = '\t'
    assign_template = 'char {identifier}[{bytesize}] = '
    pre = ''
    post = separator


class CCodeReader(CodeReader, CCode):

    @classmethod
    def decode_int(cls, cvalue):
        """Parse integer from C/Python/JS code."""
        cvalue = cvalue.strip()
        # char value is also int in C
        if len(cvalue) == 3 and cvalue[0] == cvalue[-1] == "'":
            return ord(cvalue[1])
        # C suffixes
        while cvalue[-1:].lower() in ('u', 'l'):
            cvalue = cvalue[:-1]
        if cvalue.startswith('0') and cvalue[1:2] and cvalue[1:2] in string.digits:
            # C / Python-2 octal 0777
            cvalue = '0o' + cvalue[1:]
        # 0x, 0b, decimals - like Python
        return int(cvalue, 0)

    @classmethod
    def decode_struct(cls, coded_data, fields):
        """Decode struct value from list."""
        return Props(**{
            # may be `.name = 0` or just `0`
            _key: _field.rpartition(cls.assign)[-1].strip()
            for _key, _field in zip(fields, coded_data)
        })

    @classmethod
    def decode_struct_array(cls, coded_data, fields):
        """Decode struct value from list."""
        array_len = len(coded_data) // len(fields)
        iterators = [iter(coded_data)] * len(fields)
        return tuple(
            Props(**{
                _key: _field.rpartition(cls.assign)[-1].strip()
                for _key, _field in zip(fields, _chunk)
            })
            for _chunk in zip(*iterators)
        )


class CCodeWriter(CodeWriter, CCode):

    @classmethod
    def encode_int(cls, value):
        """Output hex number in C format."""
        return f'0x{value:02x}'

    @classmethod
    def encode_struct(cls, header, fields=(), compact=False, show_names=True):
        """Encode namespace class as struct."""
        if fields:
            header_dict = {
                _name: getattr(header, _name)
                for _name in fields
            }
        else:
            header_dict = vars(header)
        if compact:
            joiner = ', '
            start, end = cls.delimiters
            indent = ''
        else:
            joiner = ',\n'
            start, end = cls.delimiters[0] + '\n', '\n' + cls.delimiters[1]
            indent = cls.indent
        if show_names:
            # format with explicit member names
            data = joiner.join(
                f'{indent}.{_name} = {_value}'
                for _name, _value in header_dict.items()
            )
            return start + data + end
        else:
            data = joiner.join(
                f'{indent}{_value}'
                for _name, _value in header_dict.items()
            )
            return start + data + end




###############################################################################
# JSON

class JSONCode:
    """JSON code."""
    delimiters = '[]'
    comment = '//'
    # JSON separator should only be written *between* multiple entries
    separator = ',\n\n'
    assign = ':'
    block_comment = ()

    indent = ' ' * 2
    assign_template = '"{identifier}": '
    pre = '{\n'
    post = '\n}\n'


class JSONCodeReader(CodeReader, JSONCode):
    decode_int = CCodeReader.decode_int

class JSONCodeWriter(CodeWriter, JSONCode):
    encode_int = CCodeWriter.encode_int


###############################################################################
# Python

class PythonCode:
    """Python source code, using lists."""
    delimiters = '[]'
    comment = '#'
    assign = '='
    block_comment = ()
    separator = '\n\n'

    indent = ' ' * 4
    assign_template = '{identifier} = '
    pre = ''
    post = separator

class PythonTupleCode(PythonCode):
    """Python source code, using tuples."""
    delimiters = '()'


class PythonCodeReader(CodeReader, PythonCode):
    # Python and C notations are not quite the same but this works
    decode_int = CCodeReader.decode_int

class PythonCodeWriter(CodeWriter, PythonCode):
    encode_int = CCodeWriter.encode_int

class PythonTupleCodeReader(CodeReader, PythonTupleCode):
    decode_int = CCodeReader.decode_int

class PythonTupleCodeWriter(CodeWriter, PythonTupleCode):
    encode_int = CCodeWriter.encode_int


###############################################################################
# Pascal

class PascalCode:
    """Pascal source code wrapper."""
    delimiters = '()'
    # pascal has block comments only
    comment = ''
    assign = '='
    block_comment = ('{','}')
    separator = ';\n\n'

    indent = ' ' * 2
    assign_template = '{identifier}: Array[1..{bytesize}] of Integer = '
    pre = 'const\n\n'
    post = separator


class PascalCodeReader(CodeReader, PascalCode):

    @classmethod
    def decode_int(cls, cvalue):
        """Parse integer from Pascal code."""
        cvalue = cvalue.strip()
        if cvalue.startswith('#'):
            # char literal
            return int(cvalue[1:], 10)
        if cvalue.startswith('$'):
            return int(cvalue[1:], 16)


class PascalCodeWriter(CodeWriter, PascalCode):

    @classmethod
    def encode_int(cls, value):
        """Output hex number in Pascal format."""
        return f'${value:02x}'



"""
monobit.storage.fontformats.raw.xbin - XBIN font section

(c) 2023--2024 Rob Hagemans
licence: https://opensource.org/licenses/MIT
"""

import logging

from monobit.storage import loaders, savers, Magic
from monobit.base import FileFormatError, UnsupportedError
from monobit.core import Glyph, Font, Char
from monobit.base.struct import little_endian as le, bitfield

from .raw import load_bitmap, save_bitmap
from monobit.storage.utils.limitations import ensure_single


###############################################################################
# XBIN font section
# https://web.archive.org/web/20120204063040/http://www.acid.org/info/xbin/x_spec.htm

_XBIN_MAGIC = b'XBIN\x1a'
_XBIN_HEADER = le.Struct(
    magic='5s',
    # Width of the image in character columns.
    width='word',
    # Height of the image in character rows.
    height='word',
    # Number of pixel rows (scanlines) in the font, Default value for VGA is 16.
    # Any value from 1 to 32 is technically possible on VGA. Any other values
    # should be considered illegal.
    fontsize='byte',
    # A set of flags indicating special features in the XBin file.
    # 7 6 5  4        3        2        1    0
    # Unused 512Chars NonBlink Compress Font Palette
    palette=bitfield('byte', 1),
    font=bitfield('byte', 1),
    compress=bitfield('byte', 1),
    nonblink=bitfield('byte', 1),
    has_512_chars=bitfield('byte', 1),
    unused_flags=bitfield('byte', 3)
)

@loaders.register(
    name='xbin',
    magic=(_XBIN_MAGIC,),
    patterns=('*.xb',),
)
def load_xbin(instream):
    """Load a XBIN font."""
    header = _XBIN_HEADER.read_from(instream)
    if header.magic != _XBIN_MAGIC:
        raise FileFormatError(
            f'Not an XBIN file: incorrect signature {header.magic}.'
        )
    if not header.font:
        raise UnsupportedError('XBIN file contains no font.')
    height = header.fontsize
    if header.has_512_chars:
        count = 512
    else:
        count = 256
    # skip 48-byte palette, if present
    if header.palette:
        instream.read(48)
    font = load_bitmap(instream, width=8, height=height, count=count)
    font = font.modify(source_format='XBIN')
    return font


@savers.register(linked=load_xbin)
def save_xbin(fonts, outstream):
    """Save an XBIN font."""
    font = ensure_single(fonts)
    codepoints = font.get_codepoints()
    if not codepoints:
        raise ValueError('No storable codepoints found in font.')
    max_cp = max(int(_cp) for _cp in codepoints)
    if max_cp >= 512:
        logging.warning('Glyphs above codepoint 512 will not be stored.')
    blank = Glyph.blank(width=8, height=font.cell_size.y)
    if max_cp >= 256:
        count = 512
    else:
        count = 256
    glyphs = (font.get_glyph(_cp, missing=blank) for _cp in range(count))
    header = _XBIN_HEADER(
        magic=_XBIN_MAGIC,
        # no image stored, so no width and height
        fontsize=font.cell_size.y,
        font=1,
        has_512_glyphs=count==512,
    )
    outstream.write(bytes(header))
    font = Font(glyphs)
    save_bitmap(outstream, font)



"""
monobit.storage.fontformats.text.dosstart - DosStart! format

(c) 2023--2024 Rob Hagemans
licence: https://opensource.org/licenses/MIT
"""

import logging
import string
from io import StringIO

from monobit.storage import loaders, savers
from monobit.base import FileFormatError, UnsupportedError
from monobit.core import Font, Glyph, Raster, StrokePath
from monobit.storage.utils.limitations import ensure_single, make_contiguous


_DOSSTART_SIG = 'DosStartFont'

@loaders.register(
    name='dosstart',
    magic=(_DOSSTART_SIG.encode('ascii'),),
    patterns=('*.dsf',),
    text=True,
)
def load_dosstart(instream):
    """Load font from DosStart! .DSF file."""
    instream = instream.text
    if instream.readline().rstrip('\r\n') != _DOSSTART_SIG:
        raise FileFormatError('Not a DosStart! font.')
    props = dict(
        name=instream.readline().rstrip('\r\n'),
    )
    try:
        format, _, ascent = instream.readline().rstrip('\r\n').partition(',')
        format = int(format)
        ascent = int(ascent)
    except ValueError as e:
        raise FileFormatError('Not a well-formed DosStart! file.')
    if format == 0:
        # ignore ascent value in format-0, shift-up is deduced from path start
        glyphs = _read_dsf_format_0(instream)
    elif format == 1:
        glyphs = _read_dsf_format_1(instream, ascent)
    else:
        raise UnsupportedError(f'Unknown DosStart font format {format}')
    return Font(glyphs, **props).label(char_from='ascii')


@savers.register(linked=load_dosstart)
def save_dosstart(fonts, outstream):
    """Save font to DosStart! .DSF file (format 1 only)."""
    font = ensure_single(fonts)
    _write_dsf_format_1(font, outstream)


###############################################################################
# DosStart! Format 1: bitmap format

def _read_dsf_format_1(instream, ascent):
    """DosStart simple bitmap format reader."""
    glyphs = []
    for i, line in enumerate(instream):
        code, _, advance = line.rstrip('\r\n').upper().partition(',')
        advance = int(advance)
        width = int(code[:3])
        height = int(code[3:6])
        raster = Raster.from_vector(
            code[6:].strip(), stride=width, inklevels='GH'
        )
        # sample file has short bit string
        raster = raster.expand(top=height-raster.height)
        glyphs.append(Glyph(
            raster, right_bearing=advance-width,
            codepoint=0x20+i, shift_up=ascent-height,
        ))
    return glyphs


def _write_dsf_format_1(font, outstream):
    """DosStart simple bitmap format writer."""
    font = font.resample(codepoints=range(0x20, 0x80), missing='empty')
    outstream.write(_DOSSTART_SIG.encode('ascii') + b'\r\n')
    outstream.write(font.name.encode('ascii', 'replace') + b'\r\n')
    outstream.write((b'1, %d' % (font.ascent,)) + b'\r\n')
    for glyph in font.glyphs:
        # code
        outstream.write(b'%03d%03d' % (glyph.width, glyph.height))
        # glyph def
        outstream.write(b''.join(glyph.as_vector(inklevels=(b'G', b'H'))))
        # advance
        outstream.write(b', %d \r\n' % (glyph.advance_width,))


###############################################################################
# DosStart! Format 0: stroke format

def _read_dsf_format_0(instream):
    """DosStart turtle graphics format."""
    glyphs = []
    for i, line in enumerate(instream):
        line = line.rstrip('\r\n')
        if not line:
            continue
        code, _, advance = line.lower().partition(',')
        advance = int(advance)
        path = []
        offset = 0
        width, height = 0, 0
        while True:
            mode, direction, offset = _read_command(code, offset)
            if not direction:
                break
            ink = mode != 'b'
            move = mode != 'n'
            step, offset = _read_stepsize(code, offset)
            dx, dy = _move_turtle(direction, step)
            pathmove = StrokePath.LINE if ink else StrokePath.MOVE
            path.append((pathmove, dx, dy))
            if not move:
                path.append((StrokePath.MOVE, -dx, -dy))
        # turtle moves to next origin, therefore raster includes this column
        glyph = Glyph.from_path(
            StrokePath(path), advance_width=advance, codepoint=0x20+i,
        )
        # remove extra column generated by path going to next origin
        glyph = glyph.crop(right=1)
        glyphs.append(glyph)
    return glyphs

def _read_command(code, offset):
    """Read mode and command characer."""
    try:
        cmd = code[offset]
    except IndexError:
        return '', '', offset
    offset += 1
    if cmd not in ('b', 'n'):
        return '', cmd, offset
    mode = cmd
    try:
        cmd = code[offset]
    except IndexError:
        cmd = ''
    offset += 1
    return mode, cmd, offset

def _read_stepsize(code, offset):
    """Read step size."""
    num = []
    while True:
        try:
            d = code[offset]
        except IndexError:
            break
        if not d.isdigit():
            break
        offset += 1
        num.append(d)
    if not num:
        return 1, offset
    return int(''.join(num)), offset

def _move_turtle(cmd, step):
    """Move the turtle one step in the given direction."""
    dx, dy = 0, 0
    if cmd in ('u', 'e', 'h'):
        dy += 1
    if cmd in ('r', 'e', 'f'):
        dx += 1
    if cmd in ('d', 'f', 'g'):
        dy -= 1
    if cmd in ('l', 'g', 'h'):
        dx -= 1
    return dx*step, dy*step



"""
monobit.storage.fontformats.fon - Windows and OS/2 FON files

(c) 2019--2024 Rob Hagemans
licence: https://opensource.org/licenses/MIT
"""

# ensure formats get registered
from . import fon, windows, os2
